import torchvision.transforms as T

from ...core.vision_encoder.tokenizer import SimpleTokenizer


def get_image_transform(
    image_size: int,
    center_crop: bool = False,
    interpolation: T.InterpolationMode = T.InterpolationMode.BILINEAR  # We used bilinear during training
):
    if center_crop:
        crop = [
            T.Resize(image_size, interpolation=interpolation),
            T.CenterCrop(image_size)
        ]
    else:
        # "Squash": most versatile
        crop = [
            T.Resize((image_size, image_size), interpolation=interpolation)
        ]
    
    return T.Compose(crop + [
        T.Lambda(lambda x: x.convert("RGB")),
        T.ToTensor(),
        T.Normalize([0.5, 0.5, 0.5], [0.5, 0.5, 0.5], inplace=True),
    ])



def get_text_tokenizer(context_length: int):
    return SimpleTokenizer(context_length=context_length)


"""
 Copyright (c) 2022, salesforce.com, inc.
 All rights reserved.
 SPDX-License-Identifier: BSD-3-Clause
 For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
"""

from dataclasses import dataclass
from typing import Optional

import torch
from transformers.modeling_outputs import (
    BaseModelOutputWithPoolingAndCrossAttentions,
    CausalLMOutputWithCrossAttentions,
    ModelOutput,
)


@dataclass
class AlbefSimilarity(ModelOutput):
    sim_i2t: torch.FloatTensor = None
    sim_t2i: torch.FloatTensor = None

    sim_i2t_m: Optional[torch.FloatTensor] = None
    sim_t2i_m: Optional[torch.FloatTensor] = None

    sim_i2t_targets: Optional[torch.FloatTensor] = None
    sim_t2i_targets: Optional[torch.FloatTensor] = None


@dataclass
class AlbefIntermediateOutput(ModelOutput):
    # uni-modal features
    image_embeds: torch.FloatTensor = None
    text_embeds: Optional[torch.FloatTensor] = None

    image_embeds_m: Optional[torch.FloatTensor] = None
    text_embeds_m: Optional[torch.FloatTensor] = None

    # intermediate outputs of multimodal encoder
    encoder_output: Optional[BaseModelOutputWithPoolingAndCrossAttentions] = None
    encoder_output_m: Optional[BaseModelOutputWithPoolingAndCrossAttentions] = None
    encoder_output_neg: Optional[BaseModelOutputWithPoolingAndCrossAttentions] = None

    itm_logits: Optional[torch.FloatTensor] = None
    itm_labels: Optional[torch.LongTensor] = None

    # intermediate outputs of multimodal decoder
    decoder_output: Optional[CausalLMOutputWithCrossAttentions] = None
    decoder_labels: Optional[torch.LongTensor] = None


@dataclass
class AlbefOutput(ModelOutput):
    # some finetuned models (e.g. BlipVQA) do not compute similarity, thus optional.
    sims: Optional[AlbefSimilarity] = None

    intermediate_output: AlbefIntermediateOutput = None

    loss: Optional[torch.FloatTensor] = None

    loss_itc: Optional[torch.FloatTensor] = None

    loss_itm: Optional[torch.FloatTensor] = None

    loss_mlm: Optional[torch.FloatTensor] = None


@dataclass
class AlbefOutputWithLogits(AlbefOutput):
    logits: torch.FloatTensor = None
    logits_m: torch.FloatTensor = None


@dataclass
class AlbefOutputFeatures(ModelOutput):
    """
    Data class of features from AlbefFeatureExtractor.

    Args:
        image_embeds: `torch.FloatTensor` of shape `(batch_size, num_patches+1, embed_dim)`, `optional`
        image_features: `torch.FloatTensor` of shape `(batch_size, num_patches+1, feature_dim)`, `optional`
        text_embeds: `torch.FloatTensor` of shape `(batch_size, sequence_length+1, embed_dim)`, `optional`
        text_features: `torch.FloatTensor` of shape `(batch_size, sequence_length+1, feature_dim)`, `optional`

        The first embedding or feature is for the [CLS] token.

        Features are obtained by projecting the corresponding embedding into a normalized low-dimensional space.
    """

    image_embeds: Optional[torch.FloatTensor] = None
    image_embeds_proj: Optional[torch.FloatTensor] = None

    text_embeds: Optional[torch.FloatTensor] = None
    text_embeds_proj: Optional[torch.FloatTensor] = None

    multimodal_embeds: Optional[torch.FloatTensor] = None



# Copyright (c) Meta Platforms, Inc. and affiliates.
# This software may be used and distributed according to the terms of the Llama 2 Community License Agreement.

import json
import logging
from collections import namedtuple
from dataclasses import asdict, dataclass
from datetime import datetime, timezone
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import torch
import torch.distributed as dist
import torch.nn as nn
import wandb

from ..core.distributed import get_is_master

Scalar = Union[int, float]

logger = logging.getLogger()


@dataclass
class WandbArgs:
    job_type: Optional[str] = None
    dir: Optional[str] = None
    project: Optional[str] = None
    entity: Optional[str] = None
    tags: Optional[List] = None
    group: Optional[str] = None
    name: Optional[str] = None
    notes: Optional[str] = None
    config_exclude_keys: Optional[List[str]] = None
    config_include_keys: Optional[List[str]] = None
    anonymous: Optional[str] = None
    mode: Optional[str] = None
    allow_val_change: Optional[bool] = None
    resume: Optional[Union[bool, str]] = None
    force: Optional[bool] = None
    tensorboard: Optional[bool] = None
    sync_tensorboard: Optional[bool] = None
    monitor_gym: Optional[bool] = None
    save_code: Optional[bool] = None
    id: Optional[str] = None
    fork_from: Optional[str] = None
    resume_from: Optional[str] = None


@dataclass
class LoggingArgs:
    freq: int = 10  # Log every freq optimizer steps
    level: str = "INFO"  # Default logging level to use
    acc_freq: Optional[int] = None  # Log every acc_freq gradient accumulation steps

    wandb: Optional[WandbArgs] = None


class MetricLogger:
    def __init__(self, outdir: Path, args: Optional[Any] = None):
        self.outdir = outdir
        self.jsonl_writer = None
        self.args = args

    def open(self):
        if self.jsonl_writer is None:
            self.jsonl_writer = open(self.outdir, "a")
        if (
            self.args is not None
            and self.args.logging.wandb is not None
            and get_is_master()
        ):
            run = wandb.init(
                config=asdict(self.args),
                **asdict(self.args.logging.wandb),
            )

    def log(self, metrics: Dict[str, Any]):
        if (
            self.args is not None
            and self.args.logging.wandb is not None
            and (wandb.run is not None)
        ):
            wandb.log(metrics, step=metrics["global_step"])

        metrics.update({"created_at": datetime.now(timezone.utc).isoformat()})
        print(json.dumps(metrics), file=self.jsonl_writer, flush=True)

    def close(self):
        if self.jsonl_writer is not None:
            self.jsonl_writer.close()
            self.jsonl_writer = None

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def __del__(self):
        self.close()


GPUMemStats = namedtuple(
    "GPUMemStats",
    [
        "max_active_gib",
        "max_active_pct",
        "max_reserved_gib",
        "max_reserved_pct",
        "num_alloc_retries",
        "num_ooms",
        "power_draw",
    ],
)


class GPUMemoryMonitor:
    """
    Class to monitor GPU memory usage
    """

    def __init__(self, device: str = "cuda:0"):
        self.device = torch.device(device)  # device object
        self.device_name = torch.cuda.get_device_name(self.device)
        self.device_index = torch.cuda.current_device()
        self.device_capacity = torch.cuda.get_device_properties(
            self.device
        ).total_memory
        self.device_capacity_gib = self._to_gib(self.device_capacity)

        # reset stats, clear cache
        torch.cuda.reset_peak_memory_stats()
        torch.cuda.empty_cache()

    def _to_gib(self, memory_in_bytes):
        # NOTE: GiB (gibibyte) is 1024, vs GB is 1000
        _gib_in_bytes = 1024 * 1024 * 1024
        memory_in_gib = memory_in_bytes / _gib_in_bytes
        return memory_in_gib

    def _to_pct(self, memory):
        return 100 * memory / self.device_capacity

    def get_peak_stats(self):
        cuda_info = torch.cuda.memory_stats(self.device)

        max_active = cuda_info["active_bytes.all.peak"]
        max_active_gib = self._to_gib(max_active)
        max_active_pct = self._to_pct(max_active)

        max_reserved = cuda_info["reserved_bytes.all.peak"]
        max_reserved_gib = self._to_gib(max_reserved)
        max_reserved_pct = self._to_pct(max_reserved)

        num_retries = cuda_info["num_alloc_retries"]
        num_ooms = cuda_info["num_ooms"]
        power_draw = torch.cuda.power_draw()

        if num_retries > 0:
            logger.warning(f"{num_retries} CUDA memory allocation retries.")
        if num_ooms > 0:
            logger.warning(f"{num_ooms} CUDA OOM errors thrown.")

        return GPUMemStats(
            max_active_gib,
            max_active_pct,
            max_reserved_gib,
            max_reserved_pct,
            num_retries,
            num_ooms,
            power_draw,
        )

    def reset_peak_stats(self):
        torch.cuda.reset_peak_memory_stats()
        torch.cuda.reset_accumulated_memory_stats()

    def __str__(self):
        mem_stats = self.get_peak_stats()
        display_str = f"{self.device_name} ({self.device_index}): {self.device_capacity_gib} GiB capacity, "
        display_str += (
            f"{mem_stats.max_reserved_gib} GiB peak, {mem_stats.max_reserved_pct}% peak"
        )
        return f"{display_str}"


def upload_train_to_wandb(
    ckpt_dir, project="perception", entity="codegen-team", train=True, eval=True
):
    import json
    from pathlib import Path

    import wandb
    from omegaconf import OmegaConf

    cfg = OmegaConf.load(Path(ckpt_dir) / "config.yaml")
    cfg = OmegaConf.to_container(cfg)

    if train:
        wandb.init(config=cfg, name=cfg["name"], project=project, entity=entity)

        with open(Path(ckpt_dir) / "metrics.jsonl") as f:
            for l in f:
                m = json.loads(l)
                wandb.log(m, step=m["global_step"])

        wandb.finish()

    if eval:
        wandb.init(config=cfg, name=cfg["name"], project=project, entity=entity)

        with open(Path(ckpt_dir) / "metrics.eval.jsonl") as f:
            for l in f:
                m = json.loads(l)
                wandb.log(
                    {
                        f"evals/{name.replace('/','.')}": value
                        for name, value in m.items()
                        if "/" in name
                    },
                    step=m["global_step"],
                )

        wandb.finish()


def get_num_params(model: nn.Module) -> int:
    """
    Get the total model params
    Args : only_trainable: whether to only count trainable params
    """
    numel = {n: p.numel() for n, p in model.named_parameters()}
    return sum(numel.values())


def log_model_params(model):
    frozen_params, unfrozen_params = [], []
    num_frozen_params, num_unfrozen_params = 0, 0

    def _fn(_model):
        num_frozen_params, num_unfrozen_params = 0, 0
        for pname, pval in _model.named_parameters():
            if pval.requires_grad:
                unfrozen_params.append(pname)
                num_unfrozen_params += pval.numel()
            else:
                frozen_params.append(pname)
                num_frozen_params += pval.numel()
        return num_frozen_params, num_unfrozen_params

    if isinstance(model, torch.nn.ModuleList):
        for m in model:
            _num_frozen_params, _num_unfrozen_params = _fn(m)
            num_frozen_params += _num_frozen_params
            num_unfrozen_params += _num_unfrozen_params
    else:
        num_frozen_params, num_unfrozen_params = _fn(model)

    logger.info(f"Logging Trainable Parameters after first step.")
    logger.debug(f"Frozen params: {frozen_params}")
    logger.debug(f"Trainable params: {unfrozen_params}")
    logger.info(
        f"Params total: {num_frozen_params + num_unfrozen_params:,}, "
        f"Learnable: {num_unfrozen_params:,}, "
        f"Frozen: {num_frozen_params:,}."
    )



import argparse
import os
from os.path import join

import pandas as pd

from training import train_for_srocc_svr
from feature_utils import conbine_features


def remove_extensions(file_name):
    if file_name.endswith('.yuv'):
        return file_name[:-4]
    elif file_name.endswith('.mp4'):
        return file_name[:-4]
    elif file_name.endswith('_4k'):
        return file_name[:-3]
    else:
        return file_name



# Main function to get features
def get_features(feats_pth):
    # Get non-linear features
    dlm_feats, vif_feats = conbine_features(feats_pth, False)
    feature = vif_feats.merge(dlm_feats, on='video')
    
    # Get linear features
    dlm_feats_non, vif_feats_non = conbine_features(feats_pth, True)
    nonlinear_features = vif_feats_non.merge(dlm_feats_non, on='video')
    
    return feature, nonlinear_features


if __name__ == "__main__":
    argparser = argparse.ArgumentParser(
        description='Train a model on a custom dataset')
    argparser.add_argument('feature_path', type=str,
                        help='Path to the folder containing the features')
    argparser.add_argument('score_csv', type=str,
                        help='Path to the score file. This should be a csv file with the following columns: video, score, content. It is critical that the video column is the same as the video name in the feature folder.')
    argparser.add_argument('--scaler_name', type=str,
                        help='The name of the Scaler', default='model_scaler.pkl')
    argparser.add_argument('--svr_name', type=str,
                        help='The name of the SVR', default='model_svr.pkl')

    args = argparser.parse_args()
    feats_pth = args.feature_path
    score_csv = args.score_csv
    scaler_name = os.path.abspath(args.scaler_name)
    svr_name = os.path.abspath(args.svr_name)

    # create the folder to save the scaler if it does not exist.
    if not os.path.exists(os.path.dirname(scaler_name)):
        os.makedirs(os.path.dirname(scaler_name))

    # create the folder to save the svr if it does not exist.
    if not os.path.exists(os.path.dirname(svr_name)):
        os.makedirs(os.path.dirname(svr_name))

    # read the score file
    scores = pd.read_csv(score_csv)
    scores['video'] = scores['video'].apply(remove_extensions)

    # Get features
    feature, nonlinear_features = get_features(feats_pth)

    # Merge features and scores
    feature = feature.merge(nonlinear_features, on='video')
    feature = feature.merge(scores[['video', 'score', 'content']], on='video')

    # Train model
    train_for_srocc_svr(feature, scaler_name, svr_name)
    print('Training finished. ')



import numpy as np

from ayase.models import DatasetStats, QualityMetrics, Sample, VideoMetadata


def test_quality_metrics_fields():
    qm = QualityMetrics()
    basic_fields = [
        "blur_score",
        "noise_score",
        "aesthetic_score",
        "technical_score",
        "motion_score",
        "camera_motion_score",
        "contrast",
        "brightness",
        "saturation",
        "clip_score",
    ]
    for field in basic_fields:
        assert hasattr(qm, field)

    for field in ["human_preference_score", "engagement_score"]:
        assert hasattr(qm, field)


def test_quality_metrics_default_values():
    qm = QualityMetrics()
    assert qm.blur_score is None
    assert qm.noise_score is None
    assert qm.aesthetic_score is None
    assert qm.technical_score is None
    assert qm.motion_score is None


def test_dataset_stats_fields():
    stats = DatasetStats(
        total_samples=100, valid_samples=90, invalid_samples=10, total_size=1000000
    )
    assert stats.avg_technical_score is None
    assert stats.avg_aesthetic_score is None
    assert stats.avg_motion_score is None
    assert stats.usability_ratio is None
    assert stats.size_distribution is None
    assert stats.duration_distribution is None


def test_quality_metrics_new_fields_exist():
    qm = QualityMetrics()
    new_fields = [
        "vmaf",
        "ms_ssim",
        "vif",
        "niqe",
        "t2v_score",
        "t2v_alignment",
        "t2v_quality",
        "dynamics_range",
        "dynamics_controllability",
        "scene_complexity",
        "compression_artifacts",
        "naturalness_score",
        "video_memorability",
        "usability_rate",
        "confidence_score",
        "hdr_quality",
        "sdr_quality",
        "temporal_information",
        "spatial_information",
        "flicker_score",
        "judder_score",
        "stutter_score",
        "dists",
        "fsim",
        "gmsd",
        "vsi_score",
        "brisque",
        "pesq_score",
        "av_sync_offset",
        "dover_score",
        "dover_technical",
        "dover_aesthetic",
        "topiq_score",
        "liqe_score",
        "clip_iqa_score",
        "color_grading_score",
        "white_balance_score",
        "exposure_consistency",
        "focus_quality",
        "banding_severity",
        "qalign_quality",
        "qalign_aesthetic",
        "pickscore_score",
        "hpsv3_score",
        "chipqa_score",
        "hdr_chipqa_score",
        "hdrmax_score",
        "brightrate_score",
        "face_quality_score",
        "face_identity_consistency",
        "face_expression_smoothness",
        "face_landmark_jitter",
        "object_permanence_score",
        "semantic_consistency",
        "depth_temporal_consistency",
        "codec_efficiency",
        "gop_quality",
        "codec_artifacts",
        "deepfake_probability",
        "harmful_content_score",
        "watermark_strength",
        "bias_score",
        "depth_quality",
        "multiview_consistency",
        "stereo_comfort_score",
        "musiq_score",
        "contrique_score",
        "mdtvsfa_score",
        ]
    for field in new_fields:
        assert hasattr(qm, field)
        assert getattr(qm, field) is None


def test_quality_metrics_accept_values():
    qm = QualityMetrics(
        flicker_score=25.0,
        brisque=42.0,
        deepfake_probability=0.1,
        musiq_score=75.0,
    )
    assert qm.flicker_score == 25.0
    assert qm.brisque == 42.0
    assert qm.deepfake_probability == 0.1
    assert qm.musiq_score == 75.0


def test_dataset_stats_new_fields():
    stats = DatasetStats(
        total_samples=100,
        valid_samples=95,
        invalid_samples=5,
        total_size=1_000_000,
    )
    new_fields = [
        "fvd",
        "kvd",
        "fvmd",
        "fid",
        "precision",
        "recall",
        "coverage",
        "density",
        "diversity_score",
        "semantic_coverage",
        "outlier_count",
        "class_balance_score",
        "duplicate_pairs",
    ]
    for field in new_fields:
        assert hasattr(stats, field)


def test_video_metadata_fields():
    metadata = VideoMetadata(
        width=1920,
        height=1080,
        fps=30.0,
        duration=10.0,
        frame_count=300,
        codec="h264",
        bitrate=5000000,
        file_size=6250000,
    )
    assert metadata.width == 1920
    assert metadata.height == 1080
    assert metadata.fps == 30.0
    assert metadata.duration == 10.0
    assert metadata.frame_count == 300
    assert metadata.codec == "h264"
    assert metadata.bitrate == 5000000
    assert metadata.file_size == 6250000


def test_sample_metadata_path_resolution(tmp_dir):
    sample = Sample(path=tmp_dir / "test.mp4", is_video=True)
    assert sample.path == tmp_dir / "test.mp4"


def test_sample_image_loading(tmp_dir):
    import cv2

    img = np.zeros((100, 100, 3), dtype=np.uint8)
    path = tmp_dir / "test.png"
    cv2.imwrite(str(path), img)
    sample = Sample(path=path, is_video=False)
    loaded = sample.load_image()
    assert loaded is not None
    assert loaded.shape == img.shape



"""OCR Fidelity — measures whether text requested in the caption appears in video frames.

Based on the EvalCrafter OCR score methodology:
1. Extract expected text from the caption (quoted strings like "Hello", 'World')
2. Run PaddleOCR on sampled video frames to recognize rendered text
3. Compute Normalized Edit Distance (NED) between expected and recognized text
4. Score = (1 - NED) * 100 → higher means text was rendered more accurately

This is fundamentally different from the ``text_detection`` module, which only
measures text *area coverage*.  This module checks text *accuracy* — whether the
video actually renders the words the prompt asked for.

References:
    - EvalCrafter (Liu et al., 2023) — T2V benchmark OCR score
    - PaddleOCR (Baidu, 2020) — open-source OCR engine
"""

import logging
import re
from typing import List, Optional

import cv2
import numpy as np

from ayase.models import QualityMetrics, Sample, ValidationIssue, ValidationSeverity
from ayase.pipeline import PipelineModule

logger = logging.getLogger(__name__)


def _extract_quoted_text(caption: str) -> List[str]:
    """Extract text enclosed in quotes from a caption.

    Supports double quotes, single quotes, guillemets, and backticks.
    Returns a list of non-empty extracted strings.
    """
    patterns = [
        r'"([^"]+)"',     # "text"
        r"'([^']+)'",     # 'text'
        r"\u00ab([^\u00bb]+)\u00bb",  # «text»
        r"\u201c([^\u201d]+)\u201d",  # \u201ctext\u201d
        r"`([^`]+)`",     # `text`
    ]
    found: List[str] = []
    for pat in patterns:
        found.extend(re.findall(pat, caption))
    return [t.strip() for t in found if t.strip()]


def _normalized_edit_distance(reference: str, hypothesis: str) -> float:
    """Compute Normalized Edit Distance (NED) between two strings.

    Returns a value in [0, 1] where 0 means identical strings.
    Uses the standard Levenshtein distance normalized by max length.
    """
    if not reference and not hypothesis:
        return 0.0
    try:
        from Levenshtein import distance as lev_distance

        d = lev_distance(reference, hypothesis)
    except ImportError:
        # Fallback: simple DP Levenshtein
        d = _levenshtein_dp(reference, hypothesis)
    return d / max(len(reference), len(hypothesis))


def _character_error_rate(reference: str, hypothesis: str) -> float:
    """Compute Character Error Rate (CER) between two strings.

    CER = Levenshtein(ref, hyp) / len(ref).
    Returns a value in [0, ∞) where 0 means identical strings.
    Clamped to [0, 1] for scoring purposes.
    """
    if not reference:
        return 0.0 if not hypothesis else 1.0
    try:
        from Levenshtein import distance as lev_distance

        d = lev_distance(reference, hypothesis)
    except ImportError:
        d = _levenshtein_dp(reference, hypothesis)
    return min(1.0, d / len(reference))


def _word_error_rate(reference: str, hypothesis: str) -> float:
    """Compute Word Error Rate (WER) between two strings.

    WER = Levenshtein(ref_words, hyp_words) / len(ref_words).
    Returns a value in [0, ∞) where 0 means identical word sequences.
    Clamped to [0, 1] for scoring purposes.
    """
    ref_words = reference.split()
    hyp_words = hypothesis.split()
    if not ref_words:
        return 0.0 if not hyp_words else 1.0
    try:
        from Levenshtein import distance as lev_distance

        d = lev_distance(ref_words, hyp_words)
    except (ImportError, TypeError):
        d = _levenshtein_dp(ref_words, hyp_words)
    return min(1.0, d / len(ref_words))


def _levenshtein_dp(s, t) -> int:
    """Minimal Levenshtein distance implementation (no external deps). Works on strings or lists."""
    m, n = len(s), len(t)
    prev = list(range(n + 1))
    for i in range(1, m + 1):
        curr = [i] + [0] * n
        for j in range(1, n + 1):
            cost = 0 if s[i - 1] == t[j - 1] else 1
            curr[j] = min(curr[j - 1] + 1, prev[j] + 1, prev[j - 1] + cost)
        prev = curr
    return prev[n]


class OCRFidelityModule(PipelineModule):
    name = "ocr_fidelity"
    description = "Checks whether text requested in the caption actually appears in video frames (EvalCrafter OCR)"
    default_config = {
        "num_frames": 8,
        "lang": "en",
    }

    def __init__(self, config=None):
        super().__init__(config)
        self.num_frames = self.config.get("num_frames", 8)
        self.lang = self.config.get("lang", "en")
        self._ocr = None
        self._ocr_available = False

    def setup(self) -> None:
        try:
            # PaddleOCR 2.x uses np.sctypes which was removed in NumPy 2.0
            import numpy as _np
            if not hasattr(_np, "sctypes"):
                _np.sctypes = {
                    "int": [_np.int8, _np.int16, _np.int32, _np.int64],
                    "uint": [_np.uint8, _np.uint16, _np.uint32, _np.uint64],
                    "float": [_np.float16, _np.float32, _np.float64],
                    "complex": [_np.complex64, _np.complex128],
                    "others": [bool, object, bytes, str, _np.void],
                }

            from paddleocr import PaddleOCR
            import paddleocr

            logger.info("Loading PaddleOCR for OCR Fidelity...")
            # PaddleOCR >=3.x removed the show_log argument
            ocr_version = getattr(paddleocr, "__version__", "0.0.0")
            major = int(ocr_version.split(".")[0])
            if major >= 3:
                self._ocr = PaddleOCR(use_angle_cls=True, lang=self.lang)
            else:
                self._ocr = PaddleOCR(use_angle_cls=True, lang=self.lang, show_log=False)
            self._ocr_available = True
        except ImportError:
            logger.warning("PaddleOCR not installed. OCR Fidelity disabled.")
        except Exception as e:
            logger.error(f"Failed to init PaddleOCR: {e}")

    def process(self, sample: Sample) -> Sample:
        if not self._ocr_available:
            return sample

        # Prefer explicit expected_text from config (set by downstream caller)
        explicit = self.config.get("expected_text")
        if explicit:
            expected_texts = [explicit] if isinstance(explicit, str) else list(explicit)
        else:
            # Fallback: extract quoted text from caption
            caption_text = self._get_caption(sample)
            if not caption_text:
                return sample
            expected_texts = _extract_quoted_text(caption_text)

        if not expected_texts:
            return sample

        expected = " ".join(expected_texts).lower()

        try:
            frames = self._load_frames(sample)
            if not frames:
                return sample

            # Run OCR on each frame, collect all recognized text
            all_recognized: List[str] = []
            for frame in frames:
                result = self._ocr.ocr(frame, cls=True)
                if result and result[0]:
                    for line in result[0]:
                        txt = line[1][0]
                        if txt:
                            all_recognized.append(txt)

            recognized = " ".join(all_recognized).lower()

            ned = _normalized_edit_distance(expected, recognized)
            cer = _character_error_rate(expected, recognized)
            wer = _word_error_rate(expected, recognized)
            score = max(0.0, (1.0 - ned)) * 100.0

            if sample.quality_metrics is None:
                sample.quality_metrics = QualityMetrics()
            sample.quality_metrics.ocr_fidelity = round(score, 2)
            sample.quality_metrics.ocr_score = round(score, 2)
            sample.quality_metrics.ocr_cer = round(cer, 4)
            sample.quality_metrics.ocr_wer = round(wer, 4)

            if score < 30.0:
                sample.validation_issues.append(
                    ValidationIssue(
                        severity=ValidationSeverity.WARNING,
                        message=f"Low OCR fidelity: {score:.1f}% — text in video doesn't match caption",
                        details={
                            "expected_text": expected,
                            "recognized_text": recognized[:200],
                            "ocr_fidelity": score,
                            "cer": round(cer, 4),
                            "wer": round(wer, 4),
                        },
                        recommendation="Video may not be rendering the requested text correctly.",
                    )
                )

        except Exception as e:
            logger.warning(f"OCR Fidelity failed for {sample.path}: {e}")

        return sample

    def _get_caption(self, sample: Sample) -> Optional[str]:
        if sample.caption and sample.caption.text:
            return sample.caption.text
        txt_path = sample.path.with_suffix(".txt")
        if txt_path.exists():
            try:
                return txt_path.read_text(encoding="utf-8").strip()
            except Exception:
                pass
        return None

    def _load_frames(self, sample: Sample) -> List[np.ndarray]:
        frames: List[np.ndarray] = []
        try:
            if sample.is_video:
                cap = cv2.VideoCapture(str(sample.path))
                total = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
                if total <= 0:
                    cap.release()
                    return frames
                n = min(self.num_frames, total)
                indices = np.linspace(0, total - 1, n, dtype=int)
                for idx in indices:
                    cap.set(cv2.CAP_PROP_POS_FRAMES, int(idx))
                    ret, frame = cap.read()
                    if ret:
                        frames.append(frame)
                cap.release()
            else:
                img = cv2.imread(str(sample.path))
                if img is not None:
                    frames.append(img)
        except Exception as e:
            logger.debug(f"Frame loading failed: {e}")
        return frames




"""VIIDEO — Video Intrinsic Integrity and Distortion Evaluation Oracle.

Mittal et al. 2016 — completely blind NR-VQA using natural video
statistics of frame differences. No training data or human opinions
required; relies on statistical regularity of natural videos.

viideo_score — LOWER = better quality (distortion measure)
"""

import logging
import cv2
import numpy as np
from typing import Optional

from ayase.models import Sample, QualityMetrics
from ayase.pipeline import PipelineModule

logger = logging.getLogger(__name__)


def _compute_nss_params(patch: np.ndarray) -> tuple:
    """Compute generalized Gaussian distribution parameters for NSS."""
    mu = np.mean(patch)
    sigma = np.std(patch) + 1e-7
    skew = float(np.mean(((patch - mu) / sigma) ** 3))
    kurt = float(np.mean(((patch - mu) / sigma) ** 4))
    return mu, sigma, skew, kurt


def _frame_difference_features(gray1: np.ndarray, gray2: np.ndarray) -> np.ndarray:
    """Compute NSS features on frame difference."""
    diff = gray2.astype(np.float64) - gray1.astype(np.float64)

    # MSCN of frame difference
    mu = cv2.GaussianBlur(diff, (7, 7), 7 / 6)
    sigma = np.sqrt(
        np.abs(cv2.GaussianBlur(diff * diff, (7, 7), 7 / 6) - mu * mu)
    )
    sigma = np.maximum(sigma, 1e-7)
    mscn = (diff - mu) / sigma

    features = []

    # Global statistics of MSCN coefficients
    features.append(np.mean(mscn))
    features.append(np.var(mscn))
    features.append(float(np.mean(np.abs(mscn))))

    # Kurtosis
    std_val = np.std(mscn) + 1e-7
    features.append(float(np.mean(((mscn - np.mean(mscn)) / std_val) ** 4)))

    # Paired product features (4 orientations)
    for shift in [(0, 1), (1, 0), (1, 1), (1, -1)]:
        shifted = np.roll(np.roll(mscn, shift[0], axis=0), shift[1], axis=1)
        paired = mscn * shifted
        features.append(np.mean(paired))
        features.append(np.var(paired))

    # Energy of difference
    features.append(np.mean(diff ** 2))

    return np.array(features, dtype=np.float64)


class VIIDEOModule(PipelineModule):
    name = "viideo"
    description = "VIIDEO blind NR-VQA via natural video statistics (Mittal 2016, lower=better)"
    default_config = {
        "subsample": 8,
    }

    def __init__(self, config=None):
        super().__init__(config)
        self.subsample = self.config.get("subsample", 8)
        self._model = None
        self._backend = "heuristic"

    def setup(self) -> None:
        # Tier 1: Try native VIIDEO package
        try:
            import viideo
            self._model = viideo
            self._backend = "native"
            logger.info("VIIDEO (native) initialised")
            return
        except ImportError:
            pass

        # Tier 2: Heuristic fallback always available
        self._backend = "heuristic"
        logger.info("VIIDEO (heuristic) initialised — install viideo for full model")

    def process(self, sample: Sample) -> Sample:
        try:
            if self._backend == "native":
                score = float(self._model.predict(str(sample.path)))
            else:
                score = self._process_heuristic(sample)

            if score is not None:
                if sample.quality_metrics is None:
                    sample.quality_metrics = QualityMetrics()
                sample.quality_metrics.viideo_score = score

        except Exception as e:
            logger.warning(f"VIIDEO failed for {sample.path}: {e}")

        return sample

    def _process_heuristic(self, sample: Sample) -> Optional[float]:
        """Heuristic: NSS features on frame differences — lower = better."""
        if not sample.is_video:
            # VIIDEO is video-only; for images return neutral score
            return 0.0

        cap = cv2.VideoCapture(str(sample.path))
        try:
            total = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
            if total <= 1:
                return None

            n_frames = min(self.subsample + 1, total)
            indices = np.linspace(0, total - 1, n_frames, dtype=int)

            frames = []
            for idx in indices:
                cap.set(cv2.CAP_PROP_POS_FRAMES, idx)
                ret, frame = cap.read()
                if ret:
                    gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY).astype(np.float64)
                    frames.append(gray)
        finally:
            cap.release()

        if len(frames) < 2:
            return None

        # Compute NSS features on consecutive frame differences
        all_features = []
        for i in range(len(frames) - 1):
            feats = _frame_difference_features(frames[i], frames[i + 1])
            all_features.append(feats)

        feat_matrix = np.array(all_features)

        # VIIDEO intrinsic measure: higher variance in NSS params = more distortion
        # Across-time variance of features indicates temporal irregularity
        temporal_var = np.mean(np.var(feat_matrix, axis=0))

        # Mean energy of frame differences (motion-compensated distortion)
        mean_energy = np.mean(feat_matrix[:, -1])  # last feature is diff energy

        # NSS regularity: deviation from expected statistics
        mean_abs_mscn = np.mean(feat_matrix[:, 2])  # mean abs MSCN
        nss_deviation = abs(mean_abs_mscn - 0.798)  # ideal GGD value ~0.798

        # Combine into distortion score (lower = better quality)
        score = (
            0.35 * min(temporal_var * 0.1, 1.0)
            + 0.30 * min(mean_energy / 2000.0, 1.0)
            + 0.35 * min(nss_deviation * 2.0, 1.0)
        )

        return float(np.clip(score, 0.0, 1.0))



"""DSG — Davidsonian Scene Graph faithfulness (ICLR 2024, Google).

Decomposes a text caption into atomic propositions (questions) and
verifies each one against the generated image/video via visual question
answering.  The final score is the fraction of questions answered
affirmatively.

Backend tiers:
  1. **dsg** — official ``dsg-t2i`` package (LLM decomposition + VQA)
  2. **heuristic** — CLIP cosine similarity between caption sub-phrases
     and the visual content as a proxy for fine-grained faithfulness

dsg_score — higher = better faithfulness (0-1)
Requires a caption (``sample.caption.text`` or ``sample.auto_caption``).
"""

import logging
import re
from typing import List, Optional

import cv2
import numpy as np

from ayase.models import Sample, QualityMetrics
from ayase.pipeline import PipelineModule

logger = logging.getLogger(__name__)


def _get_caption_text(sample: Sample) -> Optional[str]:
    """Extract caption text from sample metadata."""
    if sample.caption is not None and sample.caption.text:
        return sample.caption.text
    qm = sample.quality_metrics
    if qm is not None and qm.auto_caption:
        return qm.auto_caption
    return None


def _decompose_caption(text: str) -> List[str]:
    """Heuristic Davidsonian decomposition: split caption into atomic clauses.

    Real DSG uses an LLM to produce formal semantic decomposition.  This
    approximation splits on conjunctions, commas, and relative clauses to
    produce short sub-phrases that can each be verified independently.
    """
    # Normalise whitespace
    text = " ".join(text.split())

    # Split on sentence boundaries, conjunctions, commas, semicolons
    parts = re.split(r"[.;]\s+|\s+and\s+|\s+but\s+|\s+while\s+|,\s+(?:and|but|which|who|that|where)\s+|,\s+", text)

    # Filter out very short fragments
    clauses = [p.strip() for p in parts if len(p.strip()) > 5]

    # If nothing split, return the whole caption as a single proposition
    if not clauses:
        clauses = [text.strip()]

    return clauses


def _clip_similarity(image: np.ndarray, texts: List[str]) -> List[float]:
    """Compute CLIP cosine similarity between an image and text prompts.

    Falls back to a simplistic feature-based proxy when CLIP is unavailable.
    """
    try:
        import torch
        import clip as clip_module

        device = "cuda" if torch.cuda.is_available() else "cpu"
        model, preprocess = clip_module.load("ViT-B/32", device=device)

        from PIL import Image
        pil_img = Image.fromarray(cv2.cvtColor(image, cv2.COLOR_BGR2RGB))
        img_input = preprocess(pil_img).unsqueeze(0).to(device)
        text_tokens = clip_module.tokenize(texts, truncate=True).to(device)

        with torch.no_grad():
            img_feat = model.encode_image(img_input)
            txt_feat = model.encode_text(text_tokens)
            img_feat = img_feat / img_feat.norm(dim=-1, keepdim=True)
            txt_feat = txt_feat / txt_feat.norm(dim=-1, keepdim=True)
            sims = (img_feat @ txt_feat.T).squeeze(0).cpu().numpy().tolist()

        return sims
    except ImportError:
        pass

    # Fallback: trivially check keyword presence via simple colour/object heuristic
    # This is a very rough proxy — returns moderate scores
    scores = []
    for text in texts:
        # Longer sub-phrases get slightly lower default (harder to verify)
        word_count = len(text.split())
        base = 0.6 - 0.02 * min(word_count, 10)
        scores.append(max(0.1, base))
    return scores


class DSGModule(PipelineModule):
    name = "dsg"
    description = "DSG Davidsonian Scene Graph faithfulness (ICLR 2024, Google)"
    default_config = {
        "threshold": 0.25,  # CLIP cosine-sim threshold to consider a proposition verified
        "subsample": 4,     # frames to sample for video
    }

    def __init__(self, config=None):
        super().__init__(config)
        self._model = None
        self._backend = "heuristic"

    def setup(self) -> None:
        # Tier 1: official dsg-t2i package
        try:
            import dsg  # noqa: F401
            self._model = dsg
            self._backend = "native"
            logger.info("DSG initialised (native dsg-t2i package)")
            return
        except ImportError:
            pass

        # Tier 2: heuristic CLIP-based decomposition
        self._backend = "heuristic"
        logger.info(
            "DSG initialised (heuristic) — "
            "install dsg-t2i for full Davidsonian decomposition"
        )

    def process(self, sample: Sample) -> Sample:
        caption = _get_caption_text(sample)
        if caption is None:
            logger.debug("DSG: no caption for %s, skipping", sample.path.name)
            return sample

        try:
            if self._backend == "native":
                score = self._process_native(sample, caption)
            else:
                score = self._process_heuristic(sample, caption)

            if score is not None:
                if sample.quality_metrics is None:
                    sample.quality_metrics = QualityMetrics()
                sample.quality_metrics.dsg_score = score
        except Exception as e:
            logger.warning("DSG failed for %s: %s", sample.path, e)

        return sample

    def _process_native(self, sample: Sample, caption: str) -> Optional[float]:
        """Use the official dsg-t2i package."""
        # The official package handles decomposition + VQA internally
        result = self._model.evaluate(str(sample.path), caption)
        if isinstance(result, dict):
            return float(result.get("score", result.get("dsg_score", 0.0)))
        return float(result)

    def _process_heuristic(self, sample: Sample, caption: str) -> Optional[float]:
        """Heuristic: decompose caption → CLIP-verify each sub-phrase."""
        threshold = self.config.get("threshold", 0.25)
        subsample = self.config.get("subsample", 4)

        # Get representative frame(s)
        frame = self._get_representative_frame(sample, subsample)
        if frame is None:
            return None

        # Decompose caption into atomic propositions
        propositions = _decompose_caption(caption)
        if not propositions:
            return None

        # Score each proposition against the image
        similarities = _clip_similarity(frame, propositions)

        # Fraction of propositions that pass the threshold
        verified = sum(1 for s in similarities if s >= threshold)
        score = verified / len(propositions)

        return float(np.clip(score, 0.0, 1.0))

    def _get_representative_frame(self, sample: Sample, subsample: int) -> Optional[np.ndarray]:
        """Extract a representative frame from image or video."""
        if sample.is_video:
            cap = cv2.VideoCapture(str(sample.path))
            try:
                total = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
                if total <= 0:
                    return None
                # Pick the middle frame
                mid = total // 2
                cap.set(cv2.CAP_PROP_POS_FRAMES, mid)
                ret, frame = cap.read()
                return frame if ret else None
            finally:
                cap.release()
        else:
            return cv2.imread(str(sample.path))



"""Tests for judder_stutter module."""

from ..conftest import _test_module_basics
from ayase.models import QualityMetrics


def test_judder_stutter_basics():
    from ayase.modules.judder_stutter import JudderStutterModule
    _test_module_basics(JudderStutterModule, "judder_stutter")

def test_judder_stutter_video(video_sample):
    from ayase.modules.judder_stutter import JudderStutterModule
    video_sample.quality_metrics = QualityMetrics()
    m = JudderStutterModule()
    m.on_mount()
    result = m.process(video_sample)
    assert result is video_sample



import argparse
import os
import socket

from utils import has_slurm, random_port, runcmd

EXP_DIR_ENV_NAME = "VL_EXP_DIR"

# if key in hostname; apply the args in value to slurm.
DEFAULT_SLURM_ARGS = dict(login="-p gpu --mem=240GB -c 64 -t 2-00:00:00")


def get_default_slurm_args():
    """get the slurm args for different cluster.
    Returns: TODO

    """
    hostname = socket.gethostname()
    for k, v in DEFAULT_SLURM_ARGS.items():
        if k in hostname:
            return v
    return ""


def parse_args():
    parser = argparse.ArgumentParser()

    # slurm
    parser.add_argument("--slurm_args", type=str, default="", help="args for slurm.")
    parser.add_argument(
        "--no_slurm",
        action="store_true",
        help="If specified, will launch job without slurm",
    )
    parser.add_argument("--jobname", type=str, required=True, help="experiment name")
    parser.add_argument(
        "--dep_jobname", type=str, default="impossible_jobname", help="the dependent job name"
    )
    parser.add_argument("--nnodes", "-n", type=int, default=1, help="the number of nodes")
    parser.add_argument(
        "--ngpus", "-g", type=int, default=1, help="the number of gpus per nodes"
    )

    #
    parser.add_argument(
        "--task",
        type=str,
        required=True,
        help="one of: pretrain, retrieval, retrieval_mc, vqa.",
    )
    parser.add_argument("--config", type=str, required=True, help="config file name.")
    parser.add_argument("--model_args", type=str, default="", help="args for model")

    args = parser.parse_args()
    return args


def get_output_dir(args):
    """get the output_dir"""
    return os.path.join(os.environ[EXP_DIR_ENV_NAME], args.jobname)


def prepare(args: argparse.Namespace):
    """prepare for job submission

    Args:
        args (dict): The arguments.

    Returns: The path to the copied source code.

    """

    output_dir = get_output_dir(args)
    code_dir = os.path.join(output_dir, "code")
    project_dirname = os.path.basename(os.getcwd())

    # check output_dir exist
    if os.path.isdir(output_dir):
        # if using slurm
        if has_slurm() and not args.no_slurm:
            raise ValueError(f"output_dir {output_dir} already exist. Exit.")
    else:
        os.mkdir(output_dir)
        # copy code
        cmd = f"cd ..; rsync -ar {project_dirname} {code_dir} --exclude='*.out'"
        print(cmd)
        runcmd(cmd)
    return os.path.join(code_dir, project_dirname)


def submit_job(args: argparse.Namespace):
    """TODO: Docstring for build_job_script.

    Args:
        args (argparse.Namespace): The commandline args.

    Returns: str. The script to run.

    """
    output_dir = get_output_dir(args)
    # copy code
    code_dir = prepare(args)

    # enter in the backup code
    master_port = os.environ.get("MASTER_PORT", random_port())
    init_cmd = f" cd {code_dir}; export MASTER_PORT={master_port}; "

    if has_slurm() and not args.no_slurm:
        # prepare slurm args.
        mode = "slurm"
        default_slurm_args = get_default_slurm_args()
        bin = (
            f" sbatch --output {output_dir}/%j.out --error {output_dir}/%j.out"
            f" {default_slurm_args}"
            f" {args.slurm_args} --job-name={args.jobname} --nodes {args.nnodes} "
            f" --ntasks {args.nnodes} "
            f" --gpus-per-node={args.ngpus} "
            f" --dependency=$(squeue --noheader --format %i --name {args.dep_jobname}) "
        )
    else:
        mode = "local"
        bin = "bash "

    # build job cmd
    job_cmd = (
        f" tasks/{args.task}.py"
        f" {args.config}"
        f" output_dir {output_dir}"
        f" {args.model_args}"
    )

    cmd = (
        f" {init_cmd} {bin} "
        f" tools/submit.sh "
        f" {mode} {args.nnodes} {args.ngpus} {job_cmd} "
    )

    with open(os.path.join(output_dir, "cmd.txt"), "w") as f:
        f.write(cmd)

    print(cmd)
    runcmd(cmd)


if __name__ == "__main__":
    args = parse_args()
    submit_job(args)



from joblib import Parallel,delayed
from .utils.hdr_utils import yuv_read
import numpy as np
import cv2
import os
import scipy.ndimage
import joblib
from .save_stats import *
from numba import jit,prange
from tqdm import tqdm 

C=1
def gen_gauss_window(lw, sigma):
    sd = np.float32(sigma)
    lw = int(lw)
    weights = [0.0] * (2 * lw + 1)
    weights[lw] = 1.0
    sum = 1.0
    sd *= sd
    for ii in range(1, lw + 1):
        tmp = np.exp(-0.5 * np.float32(ii * ii) / sd)
        weights[lw + ii] = tmp
        weights[lw - ii] = tmp
        sum += 2.0 * tmp
    for ii in range(2 * lw + 1):
        weights[ii] /= sum
    return weights
def compute_image_mscn_transform(image, C=1, avg_window=None, extend_mode='constant'):
    if avg_window is None:
      avg_window = gen_gauss_window(3, 7.0/6.0)
    assert len(np.shape(image)) == 2
    h, w = np.shape(image)
    mu_image = np.zeros((h, w), dtype=np.float32)
    var_image = np.zeros((h, w), dtype=np.float32)
    image = np.array(image).astype('float32')
    scipy.ndimage.correlate1d(image, avg_window, 0, mu_image, mode=extend_mode)
    scipy.ndimage.correlate1d(mu_image, avg_window, 1, mu_image, mode=extend_mode)
    scipy.ndimage.correlate1d(image**2, avg_window, 0, var_image, mode=extend_mode)
    scipy.ndimage.correlate1d(var_image, avg_window, 1, var_image, mode=extend_mode)
    var_image = np.sqrt(np.abs(var_image - mu_image**2))
    return (image - mu_image)/(var_image + C), var_image, mu_image

def spatiotemporal_mscn(img_buffer,avg_window,extend_mode='mirror'):
    st_mean = np.zeros((img_buffer.shape))
    scipy.ndimage.correlate1d(img_buffer, avg_window, 0, st_mean, mode=extend_mode)
    return st_mean



def Y_compute_lnl(Y):
    if(len(Y.shape)==2):
        Y = np.expand_dims(Y,axis=2)

    maxY = scipy.ndimage.maximum_filter(Y,size=(17,17,1))
    minY = scipy.ndimage.minimum_filter(Y,size=(17,17,1))
    Y_scaled = -1+(Y-minY)* 2/(1e-3+maxY-minY)
    Y_transform =  np.exp(np.abs(Y_scaled)*4)-1
    Y_transform[Y_scaled<0] = -Y_transform[Y_scaled<0]
    return Y_transform



def HDR_feat(frames, bit_depth=10):
    # video meta 
    framenos = len(frames)
    height = frames[0].shape[0]
    width = frames[0].shape[1]


    st_time_length = 5

    i = 0


    X_list = []
    spatavg_list = []
    feat_sd_list =  []
    sd_list= []
    
    j=0

    i = 0
    C = 1e-3
    for framenum in tqdm(range(framenos)): 
        

    
        # read video file
        Y_pq = frames[framenum][:,:,0]
        dY_pq =  cv2.resize(Y_pq,(width//2,height//2),interpolation=cv2.INTER_CUBIC)
        
        # normalize 
        Y = Y_pq
        dY = dY_pq

        
        # apply local nonlinearity
        Y_pq_nl = np.squeeze(Y_compute_lnl(Y))
        Y_down_pq_nl = np.squeeze(Y_compute_lnl(dY))

        # find MSCN
        Y_mscn_pq_nl,_,_ = compute_image_mscn_transform(Y_pq_nl,C=0.001)
        dY_mscn_pq_nl,_,_ = compute_image_mscn_transform(Y_down_pq_nl,C=0.001)

        # find HDRMAX features
        hdrmax_fullscale = extract_subband_feats(Y_mscn_pq_nl)
        hdrmax_halfscale = extract_subband_feats(dY_mscn_pq_nl)
        hdrmax = np.concatenate((hdrmax_fullscale,hdrmax_halfscale),axis=0)

        feat_sd_list.append(hdrmax)
        spatavg_list.append(hdrmax)

        i=i+1


        # compute rolling standard deviation
        if (i>=st_time_length): 


            # compute rolling standard deviation 
            sd_feats = np.std(feat_sd_list,axis=0)
            sd_list.append(sd_feats)
            feat_sd_list = []


            i=0
    
    X1 = np.average(spatavg_list,axis=0)
    X2 = np.average(sd_list,axis=0)
    X = np.concatenate((X1,X2),axis=0)
    
    return X




"""Tests for artfid module."""

from ..conftest import _test_module_basics
from ayase.models import QualityMetrics


def test_artfid_basics():
    from ayase.modules.artfid import ArtFIDModule
    _test_module_basics(ArtFIDModule, "artfid")

def test_artfid_no_reference(image_sample):
    from ayase.modules.artfid import ArtFIDModule
    m = ArtFIDModule()
    result = m.process(image_sample)
    assert result is image_sample



# Copyright (c) Meta Platforms, Inc. and affiliates.

"""
Include all available vision encoder configurations.
"""

from dataclasses import dataclass, replace

from functools import partial
from typing import Callable, Optional, Sequence, Tuple, List

from huggingface_hub import hf_hub_download



def fetch_pe_checkpoint(name: str, path: Optional[str] = None):
    path = path or f"hf://facebook/{name}:{name}.pt"

    if path.startswith("hf://"):
        # Load from huggingface
        path = path[len("hf://"):]
        repo, file = path.split(":")

        # To count the download, config.yaml is empty
        hf_hub_download(repo_id=repo, filename="config.yaml")
        return hf_hub_download(repo_id=repo, filename=file)
    else:
        return path




@dataclass
class PEConfig:
    """ Vision Tower Config. """
    patch_size: int
    width: int
    layers: int
    heads: int
    mlp_ratio: float
    output_dim: Optional[int]

    ls_init_value: float = None
    drop_path: float = 0.0

    image_size: int = 224,
    use_abs_posemb: bool = True
    use_cls_token: bool = False
    use_rope2d: bool = True

    pool_type: str = "attn"
    attn_pooler_heads: int = 8

    use_ln_pre: bool = True
    use_ln_post: bool = True


@dataclass
class PETextConfig:
    """ Text Tower Config. """
    context_length: int
    width: int
    heads: int
    layers: int

    output_dim: int

    mlp_ratio: float = 4.0
    vocab_size: int = 49408




PE_VISION_CONFIG = {}
PE_TEXT_CONFIG = {}



#########################################
#                PE CORE                #
#########################################

PE_VISION_CONFIG["PE-Core-G14-448"] = PEConfig(
    image_size=448,
    patch_size=14,
    width=1536,
    layers=50,
    heads=16,
    mlp_ratio=8960 / 1536,
    pool_type="attn",
    output_dim=1280,
    use_cls_token=False,
)
PE_TEXT_CONFIG["PE-Core-G14-448"] = PETextConfig(
    context_length=72,
    width=1280,
    heads=20,
    layers=24,
    output_dim=1280
)


PE_VISION_CONFIG["PE-Core-L14-336"] = PEConfig(
    image_size=336,
    patch_size=14,
    width=1024,
    layers=24,
    heads=16,
    mlp_ratio=4.0,
    pool_type="attn",
    output_dim=1024,
    use_cls_token=True,
)
PE_TEXT_CONFIG["PE-Core-L14-336"] = PETextConfig(
    context_length=32,
    width=1024,
    heads=16,
    layers=24,
    output_dim=1024
)


PE_VISION_CONFIG["PE-Core-B16-224"] = PEConfig(
    image_size=224,
    patch_size=16,
    width=768,
    layers=12,
    heads=12,
    mlp_ratio=4.0,
    pool_type="attn",
    output_dim=1024,
    use_cls_token=True,
)
PE_TEXT_CONFIG["PE-Core-B16-224"] = PE_TEXT_CONFIG["PE-Core-L14-336"]








#########################################
#                PE Lang                #
#########################################

PE_VISION_CONFIG["PE-Lang-G14-448"] = replace(
    PE_VISION_CONFIG["PE-Core-G14-448"],
    image_size=448,
    pool_type="none",
    use_ln_post=False,
    output_dim=None,
    ls_init_value=0.1,
    layers=47,
)

PE_VISION_CONFIG["PE-Lang-L14-448"] = replace(
    PE_VISION_CONFIG["PE-Core-L14-336"],
    image_size=448,
    pool_type="none",
    use_ln_post=False,
    output_dim=None,
    ls_init_value=0.1,
    layers=23
)



#########################################
#               PE Spatial              #
#########################################

PE_VISION_CONFIG["PE-Spatial-G14-448"] = replace(
    PE_VISION_CONFIG["PE-Core-G14-448"],
    image_size=448,
    pool_type="none",
    use_ln_post=False,
    output_dim=None,
    ls_init_value=0.1,
)



"""T2VEval — Text-to-Video Generated Video Evaluation (2025). t2veval_score — higher = better"""
import logging, cv2, numpy as np
from ayase.models import Sample, QualityMetrics
from ayase.pipeline import PipelineModule
logger = logging.getLogger(__name__)
class T2VEvalModule(PipelineModule):
    name = "t2veval"; description = "T2VEval text-video consistency+realness (2025)"; default_config = {"subsample": 8}
    def __init__(self, c=None):
        super().__init__(c); self.subsample = self.config.get("subsample", 8); self._backend = "heuristic"
    def setup(self):
        try: import t2veval; self._model = t2veval; self._backend = "native"; return
        except ImportError: pass
        self._backend = "heuristic"
    def process(self, sample):
        try:
            frames = self._ex(sample)
            if not frames: return sample
            # Technical quality
            tech = float(np.mean([min(cv2.Laplacian(cv2.cvtColor(f,cv2.COLOR_BGR2GRAY).astype(np.float64),cv2.CV_64F).var()/500,1) for f in frames]))
            # Realness: naturalness via NSS
            real_scores = []
            for f in frames:
                gray = cv2.cvtColor(f, cv2.COLOR_BGR2GRAY).astype(np.float64)
                mu = cv2.GaussianBlur(gray, (7,7), 7/6)
                sigma = np.sqrt(np.abs(cv2.GaussianBlur(gray**2,(7,7),7/6)-mu**2)+1e-7)
                mscn = (gray-mu)/sigma
                real_scores.append(1.0/(1.0+abs(np.var(mscn)-1.0)))
            realness = float(np.mean(real_scores))
            # Temporal
            t = 1.0
            if len(frames) > 1:
                d = [np.mean(np.abs(cv2.resize(cv2.cvtColor(frames[i],cv2.COLOR_BGR2GRAY).astype(float),(160,120))-cv2.resize(cv2.cvtColor(frames[i+1],cv2.COLOR_BGR2GRAY).astype(float),(160,120)))) for i in range(len(frames)-1)]
                t = 1.0/(1.0+np.var(d)*0.01)
            score = 0.35*tech + 0.35*realness + 0.30*t
            if sample.quality_metrics is None: sample.quality_metrics = QualityMetrics()
            sample.quality_metrics.t2veval_score = float(np.clip(score, 0, 1))
        except Exception as e: logger.warning(f"T2VEval failed: {e}")
        return sample
    def _ex(self, sample):
        frames=[]
        if sample.is_video:
            cap=cv2.VideoCapture(str(sample.path)); tot=int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
            if tot<=0: cap.release(); return []
            for i in np.linspace(0,tot-1,min(self.subsample,tot),dtype=int):
                cap.set(cv2.CAP_PROP_POS_FRAMES,i); r,f=cap.read()
                if r: frames.append(f)
            cap.release()
        else:
            img=cv2.imread(str(sample.path))
            if img is not None: frames.append(img)
        return frames



from ..utils.distributed import is_main_process, get_rank, get_world_size
import logging
import torch.distributed as dist
import torch
import io
import os
import json
import re
import random
import numpy as np
from os.path import join
from tqdm import trange
from PIL import Image
from PIL import ImageFile
from torchvision.transforms import PILToTensor
import librosa
import torchaudio
# import soundfile as sf

ImageFile.LOAD_TRUNCATED_IMAGES = True
Image.MAX_IMAGE_PIXELS = None

logger = logging.getLogger(__name__)


def load_audio_from_path(audio_path, client, sr, audio_reader_type, max_length=0):
    # print(f"audio_path: {audio_path}, client: {client}, sr: {sr}, audio_reader_type: {audio_reader_type}")
    if "s3://" in audio_path and client is not None:
        audio_bytes = client.get(audio_path)
        buff = io.BytesIO(audio_bytes)
    else:
        buff = audio_path
    if audio_reader_type == 'librosa':
        audio, _ = librosa.load(buff, sr=sr)
        audio = torch.from_numpy(audio)
        # audio = normalize(audio)        # normalize waveform to -1,1 due to specified sr in librosa.load
    # elif audio_reader_type == 'soundfile':
    #     audio, _ = sf.read(buff, sr=sr)
    #     audio = torch.from_numpy(audio)
    elif audio_reader_type == 'torchaudio':
        torchaudio.set_audio_backend('soundfile')   # for flac files
        audio, csr = torchaudio.load(buff)
        if csr != sr:
            trans = torchaudio.transforms.Resample(csr, sr)
            audio = trans(audio)
        if audio.size(0) == 2:
            audio = torch.mean(audio, dim=0, keepdim=False)
    else:
        raise NotImplementedError
    if max_length != 0:
    # if audio length is longer than max_length, we randomly crop it to uta length
        if audio.shape[0] >= max_length:
            max_start = audio.shape[0] - max_length
            start = random.randint(0, max_start)
            audio = audio[start: start + max_length]
            # padding = torch.zeros(audio.shape).long()
        else:
            # padding = torch.cat((torch.zeros(audio.shape), torch.ones(max_length-audio.shape[0])), -1).long()
            audio = torch.nn.functional.pad(audio, (0, max_length-audio.shape[-1]), 'constant')
    # print(f"post audio max: {audio.max()}, audio min: {audio.min()}, audio shape: {audio.shape}")
    if len(audio.shape) == 1:
        audio = audio.unsqueeze(0)
    fbank = audio * 2 ** 15
    fbank = torchaudio.compliance.kaldi.fbank(fbank, num_mel_bins=64, sample_frequency=16000, frame_length=25, frame_shift=10)
    fbank_mean = 15.41663
    fbank_std = 6.55582
    fbank = (fbank - fbank_mean) / (fbank_std * 2) # 998, 64
    return fbank


def load_image_from_path(image_path, client):
    if "s3://" in image_path and client is not None:
        value = client.Get(image_path)
        if value is None:
            logger.warning(f"Failed to load {image_path}")
        img_bytes = np.frombuffer(value, dtype=np.uint8)
        buff = io.BytesIO(img_bytes)
        image = Image.open(buff).convert('RGB')
    else:
        image = Image.open(image_path).convert('RGB')  # PIL Image
    image = PILToTensor()(image).unsqueeze(0)  # (1, C, H, W), torch.uint8
    return image


def load_anno(ann_file_list):
    """[summary]

    Args:
        ann_file_list (List[List[str, str]] or List[str, str]):
            the latter will be automatically converted to the former.
            Each sublist contains [anno_path, image_root], (or [anno_path, video_root, 'video'])
            which specifies the data type, video or image

    Returns:
        List(dict): each dict is {
            image: str or List[str],  # image_path,
            caption: str or List[str]  # caption text string
        }
    """
    if isinstance(ann_file_list, dict):
        ann_file_list = [ann_file_list]

    ann = []
    for d in ann_file_list:        

        data_root = d.data_root
        data_root_prefix = d.get("data_root_prefix", "")
        fp = d.anno_path

        cur_ann = json.load(open(fp, "r"))
        iterator = trange(len(cur_ann), desc=f"Loading {fp}") \
            if is_main_process() else range(len(cur_ann))
        for idx in iterator:
            if d.media_type == "image":
                key = "image"
            elif d.media_type in ["video", "audio_video"]:
                key = "video"
            elif d.media_type == "audio":
                key = "audio"
            else:
                raise NotImplementedError(key)
            
            # unified to have the same key for data path
            if isinstance(cur_ann[idx][key], str):
                cur_ann[idx]["image"] = data_root_prefix + join(data_root, cur_ann[idx][key])
            else:  # list
                cur_ann[idx]["image"] = [data_root_prefix + join(data_root, e) for e in cur_ann[idx][key]]
        ann += cur_ann
    return ann


def pre_text(text, max_l=None):
    assert type(text) is str, text
    text = re.sub(r"([,.'!?\"()*#:;~])", '', text.lower())
    text = text.replace('-', ' ').replace('/', ' ').replace('<person>', 'person')

    text = re.sub(r"\s{2,}", ' ', text)
    text = text.rstrip('\n').strip(' ')

    if max_l:  # truncate
        words = text.split(' ')
        if len(words) > max_l:
            text = ' '.join(words[:max_l])
    return text


def collect_result(result, result_dir, filename, is_json=True, is_list=True):
    if is_json:
        result_file = os.path.join(
            result_dir, '%s_rank%d.json' % (filename, get_rank()))
        final_result_file = os.path.join(result_dir, '%s.json' % filename)
        json.dump(result, open(result_file, 'w'))
    else:
        result_file = os.path.join(
            result_dir, '%s_rank%d.pth' % (filename, get_rank()))
        final_result_file = os.path.join(result_dir, '%s.pth' % filename)
        torch.save(result, result_file)

    dist.barrier()

    result = None
    if is_main_process():
        # combine results from all processes
        if is_list:
            result = []
        else:
            result = {}
        for rank in range(get_world_size()):
            if is_json:
                result_file = os.path.join(
                    result_dir, '%s_rank%d.json' % (filename, rank))
                res = json.load(open(result_file, 'r'))
            else:
                result_file = os.path.join(
                    result_dir, '%s_rank%d.pth' % (filename, rank))
                res = torch.load(result_file)
            if is_list:
                result += res
            else:
                result.update(res)

    return result


def sync_save_result(result, result_dir, filename, is_json=True, is_list=True):
    """gather results from multiple GPUs"""
    if is_json:
        result_file = os.path.join(
            result_dir, "dist_res", '%s_rank%d.json' % (filename, get_rank()))
        final_result_file = os.path.join(result_dir, '%s.json' % filename)
        os.makedirs(os.path.dirname(result_file), exist_ok=True)
        json.dump(result, open(result_file, 'w'))
    else:
        result_file = os.path.join(
            result_dir, "dist_res", '%s_rank%d.pth' % (filename, get_rank()))
        os.makedirs(os.path.dirname(result_file), exist_ok=True)
        final_result_file = os.path.join(result_dir, '%s.pth' % filename)
        torch.save(result, result_file)

    dist.barrier()

    if is_main_process():
        # combine results from all processes
        if is_list:
            result = []
        else:
            result = {}
        for rank in range(get_world_size()):
            if is_json:
                result_file = os.path.join(
                    result_dir, "dist_res", '%s_rank%d.json' % (filename, rank))
                res = json.load(open(result_file, 'r'))
            else:
                result_file = os.path.join(
                    result_dir, "dist_res", '%s_rank%d.pth' % (filename, rank))
                res = torch.load(result_file)
            if is_list:
                result += res
            else:
                result.update(res)
        if is_json:
            json.dump(result, open(final_result_file, 'w'))
        else:
            torch.save(result, final_result_file)

        logger.info('result file saved to %s' % final_result_file)
    dist.barrier()
    return final_result_file, result


def pad_sequences_1d(sequences, dtype=torch.long, device=torch.device("cpu"), fixed_length=None):
    """ Pad a single-nested list or a sequence of n-d array (torch.tensor or np.ndarray)
    into a (n+1)-d array, only allow the first dim has variable lengths.
    Args:
        sequences: list(n-d tensor or list)
        dtype: np.dtype or torch.dtype
        device:
        fixed_length: pad all seq in sequences to fixed length. All seq should have a length <= fixed_length.
            return will be of shape [len(sequences), fixed_length, ...]
    Returns:
        padded_seqs: ((n+1)-d tensor) padded with zeros
        mask: (2d tensor) of the same shape as the first two dims of padded_seqs,
              1 indicate valid, 0 otherwise
    Examples:
        >>> test_data_list = [[1,2,3], [1,2], [3,4,7,9]]
        >>> pad_sequences_1d(test_data_list, dtype=torch.long)
        >>> test_data_3d = [torch.randn(2,3,4), torch.randn(4,3,4), torch.randn(1,3,4)]
        >>> pad_sequences_1d(test_data_3d, dtype=torch.float)
        >>> test_data_list = [[1,2,3], [1,2], [3,4,7,9]]
        >>> pad_sequences_1d(test_data_list, dtype=np.float32)
        >>> test_data_3d = [np.random.randn(2,3,4), np.random.randn(4,3,4), np.random.randn(1,3,4)]
        >>> pad_sequences_1d(test_data_3d, dtype=np.float32)
    """
    if isinstance(sequences[0], list):
        if "torch" in str(dtype):
            sequences = [torch.tensor(s, dtype=dtype, device=device) for s in sequences]
        else:
            sequences = [np.asarray(s, dtype=dtype) for s in sequences]

    extra_dims = sequences[0].shape[1:]  # the extra dims should be the same for all elements
    lengths = [len(seq) for seq in sequences]
    if fixed_length is not None:
        max_length = fixed_length
    else:
        max_length = max(lengths)
    if isinstance(sequences[0], torch.Tensor):
        assert "torch" in str(dtype), "dtype and input type does not match"
        padded_seqs = torch.zeros((len(sequences), max_length) + extra_dims, dtype=dtype, device=device)
        mask = torch.zeros((len(sequences), max_length), dtype=torch.float32, device=device)
    else:  # np
        assert "numpy" in str(dtype), "dtype and input type does not match"
        padded_seqs = np.zeros((len(sequences), max_length) + extra_dims, dtype=dtype)
        mask = np.zeros((len(sequences), max_length), dtype=np.float32)

    for idx, seq in enumerate(sequences):
        end = lengths[idx]
        padded_seqs[idx, :end] = seq
        mask[idx, :end] = 1
    return padded_seqs, mask  # , lengths


"""Image LPIPS — perceptual distance between image pairs and diversity metric.

Computes the Learned Perceptual Image Patch Similarity (LPIPS) between a sample
image and its reference.  Also computes a dataset-level diversity metric via
``post_process()`` by averaging pairwise LPIPS across a random subset of image
pairs.

Outputs:
    image_lpips      — per-sample LPIPS distance vs reference (0-1, lower=more similar)
    lpips_diversity  — dataset-level average pairwise LPIPS (higher=more diverse)

Requires ``sample.reference_path`` for per-sample LPIPS.

Tiered backends:
    1. **lpips** library (AlexNet, VGG, or SqueezeNet backbone)
    2. **Heuristic** — SSIM-based proxy via OpenCV
"""

import logging
import random
from typing import Dict, List, Optional, Tuple

import cv2
import numpy as np

from ayase.models import QualityMetrics, Sample
from ayase.pipeline import PipelineModule

logger = logging.getLogger(__name__)


class ImageLPIPSModule(PipelineModule):
    name = "image_lpips"
    description = "LPIPS perceptual distance between image pairs and diversity metric"
    default_config = {
        "net": "alex",  # "alex", "vgg", "squeeze"
        "resize": 256,  # Resize images before computing LPIPS
        "diversity_max_pairs": 500,  # Max pairs for diversity computation
    }

    def __init__(self, config: Optional[dict] = None) -> None:
        super().__init__(config)
        self._ml_available = False
        self._backend = "heuristic"
        self._lpips_model = None
        self._device = None
        # Cache tensors for diversity computation
        self._tensor_cache: List[Tuple[str, object]] = []

    def setup(self) -> None:
        """Load LPIPS model with tiered fallback."""
        # Tier 1: Real LPIPS model
        try:
            import torch
            import lpips

            net = self.config.get("net", "alex")
            device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
            self._lpips_model = lpips.LPIPS(net=net).to(device)
            self._lpips_model.eval()
            self._device = device
            self._backend = "lpips"
            self._ml_available = True
            logger.info("ImageLPIPS: loaded LPIPS-%s on %s", net, device)
            return
        except (ImportError, Exception) as e:
            logger.info("LPIPS library unavailable: %s", e)

        # Tier 2: Heuristic (SSIM-based proxy)
        self._backend = "heuristic"
        self._ml_available = True
        logger.info("ImageLPIPS: using SSIM-based heuristic fallback")

    def process(self, sample: Sample) -> Sample:
        if sample.quality_metrics is None:
            sample.quality_metrics = QualityMetrics()

        if not self._ml_available:
            return sample

        try:
            # Load sample image
            sample_img = self._load_image(sample.path)
            if sample_img is None:
                return sample

            # Cache for diversity computation (store path + resized image)
            self._cache_for_diversity(str(sample.path), sample_img)

            # Per-sample FR LPIPS requires reference_path
            ref_path = getattr(sample, "reference_path", None)
            if ref_path is None:
                return sample

            ref_img = self._load_image(ref_path)
            if ref_img is None:
                return sample

            # Compute LPIPS distance
            distance = self._compute_distance(sample_img, ref_img)
            if distance is not None:
                sample.quality_metrics.image_lpips = float(np.clip(distance, 0.0, 1.0))

        except Exception as e:
            logger.warning("ImageLPIPS failed for %s: %s", sample.path, e)

        return sample

    def post_process(self, all_samples: List[Sample]) -> None:
        """Compute dataset-level LPIPS diversity from cached tensors."""
        if len(self._tensor_cache) < 2:
            self._tensor_cache = []
            return

        try:
            max_pairs = self.config.get("diversity_max_pairs", 500)
            n = len(self._tensor_cache)

            # Generate all possible pair indices
            all_pairs = [(i, j) for i in range(n) for j in range(i + 1, n)]

            # Subsample if too many pairs
            if len(all_pairs) > max_pairs:
                pairs = random.sample(all_pairs, max_pairs)
            else:
                pairs = all_pairs

            distances = []
            for i, j in pairs:
                _, img_a = self._tensor_cache[i]
                _, img_b = self._tensor_cache[j]
                dist = self._compute_distance(img_a, img_b)
                if dist is not None:
                    distances.append(dist)

            if distances:
                diversity = float(np.mean(distances))
                # Store in pipeline stats
                if hasattr(self, "pipeline") and self.pipeline:
                    if hasattr(self.pipeline, "add_dataset_metric"):
                        self.pipeline.add_dataset_metric("lpips_diversity", diversity)
                logger.info(
                    "ImageLPIPS diversity: %.4f (from %d pairs)", diversity, len(distances)
                )

        except Exception as e:
            logger.warning("ImageLPIPS diversity computation failed: %s", e)
        finally:
            self._tensor_cache = []

    # -- Internal methods -------------------------------------------------------

    def _load_image(self, path) -> Optional[np.ndarray]:
        """Load and resize an image to the configured size."""
        try:
            img = cv2.imread(str(path))
            if img is None:
                return None
            resize = self.config.get("resize", 256)
            img = cv2.resize(img, (resize, resize))
            rgb = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
            return rgb
        except Exception:
            return None

    def _cache_for_diversity(self, path: str, img: np.ndarray) -> None:
        """Cache an image for diversity computation."""
        self._tensor_cache.append((path, img))

    def _compute_distance(self, img_a: np.ndarray, img_b: np.ndarray) -> Optional[float]:
        """Compute perceptual distance between two RGB images using best backend."""
        if self._backend == "lpips":
            return self._compute_lpips(img_a, img_b)
        return self._compute_heuristic(img_a, img_b)

    def _compute_lpips(self, img_a: np.ndarray, img_b: np.ndarray) -> Optional[float]:
        """Compute distance using real LPIPS model."""
        try:
            import torch

            # Convert to tensors: [1, 3, H, W] in [-1, 1]
            t_a = torch.from_numpy(img_a).permute(2, 0, 1).unsqueeze(0).float() / 127.5 - 1.0
            t_b = torch.from_numpy(img_b).permute(2, 0, 1).unsqueeze(0).float() / 127.5 - 1.0
            t_a, t_b = t_a.to(self._device), t_b.to(self._device)

            with torch.no_grad():
                dist = self._lpips_model(t_a, t_b).item()
            return float(dist)
        except Exception as e:
            logger.debug("LPIPS computation failed: %s", e)
            return None

    def _compute_heuristic(self, img_a: np.ndarray, img_b: np.ndarray) -> Optional[float]:
        """SSIM-based perceptual distance proxy.

        Converts SSIM (1=identical, -1=different) to a distance in [0, 1].
        """
        try:
            gray_a = cv2.cvtColor(img_a, cv2.COLOR_RGB2GRAY).astype(np.float64)
            gray_b = cv2.cvtColor(img_b, cv2.COLOR_RGB2GRAY).astype(np.float64)

            # Compute SSIM manually
            c1 = (0.01 * 255) ** 2
            c2 = (0.03 * 255) ** 2

            mu_a = cv2.GaussianBlur(gray_a, (11, 11), 1.5)
            mu_b = cv2.GaussianBlur(gray_b, (11, 11), 1.5)

            mu_a_sq = mu_a ** 2
            mu_b_sq = mu_b ** 2
            mu_ab = mu_a * mu_b

            sigma_a_sq = cv2.GaussianBlur(gray_a ** 2, (11, 11), 1.5) - mu_a_sq
            sigma_b_sq = cv2.GaussianBlur(gray_b ** 2, (11, 11), 1.5) - mu_b_sq
            sigma_ab = cv2.GaussianBlur(gray_a * gray_b, (11, 11), 1.5) - mu_ab

            ssim_map = ((2 * mu_ab + c1) * (2 * sigma_ab + c2)) / (
                (mu_a_sq + mu_b_sq + c1) * (sigma_a_sq + sigma_b_sq + c2)
            )
            ssim_val = float(np.mean(ssim_map))

            # Convert SSIM to distance: SSIM=1 -> dist=0, SSIM=0 -> dist=0.5
            distance = float(np.clip((1.0 - ssim_val) / 2.0, 0.0, 1.0))
            return distance
        except Exception as e:
            logger.debug("Heuristic SSIM distance failed: %s", e)
            return None



# Copyright (c) Meta Platforms, Inc. and affiliates.

import logging
import os
from typing import Any, Callable, Dict, List, Optional

import numpy as np
import torch
from PIL import Image

from ...apps.plm.dataset_conf import DatasetConf

logger = logging.getLogger(__name__)


class VisionPreprocessor:
    def __init__(
        self,
        transform: Optional[Callable],
        tokenizer: Callable,
        max_video_frames: Optional[int],
        dataset_config: DatasetConf,
    ):
        self.mllm_tokenizer = tokenizer
        self.transform = transform
        self.root_dir = ""
        if dataset_config.root_dir:
            self.root_dir = dataset_config.root_dir
        self.max_video_frames = max_video_frames

    def __call__(self, row: Dict[str, Any], rng: np.random.RandomState):
        try:
            return self.process(row, rng)
        except Exception as e:
            logger.error(f"Error processing row: {e}")
            return None  # None will be skipped in training

    def process(self, row: Dict[str, Any], rng: np.random.RandomState):
        del rng
        if "conversations" in row:
            conversations = row["conversations"]
        else:
            conversations = self.get_conversation(caption=row["text"], prompt="")

        if "bbox" in row:
            assert (
                "width" in row and "height" in row
            ), f"bbox is present in the annotation, however image width or height is not specified, which is not expected."
            w, h = row["width"], row["height"]
            bboxes = row["bbox"]
            conversations = self.transform["region"](conversations, bboxes, w, h)

        media = None
        media_type = ""
        if "image" in row:
            processed_images = []
            image_files = row["image"]
            if isinstance(image_files, str):
                image_files = [image_files]
            pil_images = []
            for image_file in image_files:
                if self.root_dir:
                    image_file = os.path.join(self.root_dir, image_file)
                try:
                    image = Image.open(image_file).convert("RGB")
                    pil_images.append(image)
                except Exception as e:
                    logger.info(
                        f"loading image failed because of the following error:\n {e}"
                    )
                    return None
            if self.transform:
                if len(pil_images) == 1:
                    transform = self.transform["image"]
                    processed_images, _ = transform(pil_images[0])
                else:
                    transform = self.transform["video"]
                    processed_images, _ = transform._process_multiple_images_pil(
                        pil_images
                    )
            if len(processed_images.shape) == 3:
                processed_images = processed_images.unsqueeze(0)
            media = processed_images
            media_type = "multi_image" if len(image_files) > 1 else "image"
        elif "video" in row:
            video_file = row["video"]
            start_time = row.get("start_time", None)
            bbox_map = row.get("bbox_map", None)
            end_time = row.get("end_time", None)
            if self.root_dir:
                video_file = os.path.join(self.root_dir, video_file)
            video_info = (
                video_file,
                self.max_video_frames,
                start_time,
                end_time,
                bbox_map,
            )
            video, _ = self.transform["video"](video_info)
            media = video
            media_type = "video"
        else:
            # This is a text-only sample. We create a dummy white image to facilitate batch processing.
            # Note that this image serves solely as a placeholder and is never used as input to the VLM.
            media = torch.ones(
                1, 3, self.transform["image"].size, self.transform["image"].size
            )
            media_type = "text"

        tokenized_sample = self.mllm_tokenizer(
            conversations=conversations, media=media, media_type=media_type
        )
        out = (
            {
                "media": media,
                "text_ids": tokenized_sample.text_ids,
                "response_pos": tokenized_sample.response_pos,
                "image_pos": tokenized_sample.image_pos,
                "num_image_chunks": tokenized_sample.num_media_chunks,
                "media_type": media_type,
            }
            if tokenized_sample.is_valid
            else None
        )  # None will be skipped in training
        return out

    def get_conversation(
        self, caption: str, prompt: str = None
    ) -> List[Dict[str, str]]:
        """
        Converts plain caption to conversation.

        Args:
            caption (str): plain caption

        Returns:
            List[Dict[str, str]]: conversation
        """
        conversations = [
            {"from": "human", "value": prompt if prompt is not None else ""},
            {"from": "assistant", "value": caption},
        ]
        return conversations



"""ArtFID — Artistic Style Transfer FID (2022).

Full-reference metric for evaluating style transfer quality.
Combines content fidelity (LPIPS-like) with style similarity (FID-like).

pip install art-fid

artfid_score — lower = better (combined content + style distance).
"""

import logging
import cv2
import numpy as np
from pathlib import Path
from typing import Optional

from ayase.models import Sample, QualityMetrics
from ayase.base_modules import ReferenceBasedModule

logger = logging.getLogger(__name__)


class ArtFIDModule(ReferenceBasedModule):
    name = "artfid"
    description = "ArtFID style transfer quality (FR, 2022, lower=better)"
    metric_field = "artfid_score"
    default_config = {"subsample": 8}

    def __init__(self, config=None):
        super().__init__(config)
        self._model = None
        self._ml_available = False
        self.subsample = self.config.get("subsample", 8)

    def setup(self) -> None:
        # Tier 1: art-fid package
        try:
            import art_fid
            self._ml_available = True
            self._model = art_fid
            logger.info("ArtFID module initialised (art-fid package)")
            return
        except ImportError:
            pass
        except Exception as e:
            logger.debug(f"art-fid init failed: {e}")

        # Tier 2: heuristic
        logger.info("ArtFID module initialised (heuristic fallback)")

    def compute_reference_score(self, sample_path: Path, reference_path: Path) -> Optional[float]:
        try:
            if str(sample_path).lower().endswith((".mp4", ".avi", ".mov", ".mkv", ".webm")):
                return self._score_video(str(sample_path), str(reference_path))
            else:
                return self._score_image(str(sample_path), str(reference_path))
        except Exception as e:
            logger.warning(f"ArtFID failed: {e}")
            return None

    def _score_image(self, sample_p: str, ref_p: str) -> Optional[float]:
        img = cv2.imread(sample_p)
        ref = cv2.imread(ref_p)
        if img is None or ref is None:
            return None
        return self._heuristic_artfid(img, ref)

    def _score_video(self, sample_p: str, ref_p: str) -> Optional[float]:
        cap_s = cv2.VideoCapture(sample_p)
        cap_r = cv2.VideoCapture(ref_p)
        try:
            total = min(
                int(cap_s.get(cv2.CAP_PROP_FRAME_COUNT)),
                int(cap_r.get(cv2.CAP_PROP_FRAME_COUNT)),
            )
            if total <= 0:
                return None
            indices = np.linspace(0, total - 1, min(self.subsample, total), dtype=int)
            scores = []
            for idx in indices:
                cap_s.set(cv2.CAP_PROP_POS_FRAMES, idx)
                cap_r.set(cv2.CAP_PROP_POS_FRAMES, idx)
                ret_s, frame_s = cap_s.read()
                ret_r, frame_r = cap_r.read()
                if ret_s and ret_r:
                    s = self._heuristic_artfid(frame_s, frame_r)
                    if s is not None:
                        scores.append(s)
            return float(np.mean(scores)) if scores else None
        finally:
            cap_s.release()
            cap_r.release()

    def _heuristic_artfid(self, img: np.ndarray, ref: np.ndarray) -> float:
        """Heuristic: content LPIPS proxy + style Gram distance proxy.

        Content term: multi-scale structural difference (like LPIPS).
        Style term: Gram matrix distance of filter responses (like style FID).
        """
        h, w = ref.shape[:2]
        img = cv2.resize(img, (w, h)).astype(np.float64) / 255.0
        ref = ref.astype(np.float64) / 255.0

        # Content distance: multi-scale MSE
        content_dists = []
        img_s, ref_s = img, ref
        for _ in range(3):
            mse = np.mean((img_s - ref_s) ** 2)
            content_dists.append(mse)
            new_h, new_w = max(img_s.shape[0] // 2, 1), max(img_s.shape[1] // 2, 1)
            img_s = cv2.resize(img_s, (new_w, new_h))
            ref_s = cv2.resize(ref_s, (new_w, new_h))
        content_dist = float(np.mean(content_dists))

        # Style distance: Gram matrix of filter bank responses
        gray_img = cv2.cvtColor((img * 255).astype(np.uint8), cv2.COLOR_BGR2GRAY).astype(np.float64)
        gray_ref = cv2.cvtColor((ref * 255).astype(np.uint8), cv2.COLOR_BGR2GRAY).astype(np.float64)

        # Simple filter bank: Sobel x, Sobel y, Laplacian
        filters_img = []
        filters_ref = []
        for fn in [
            lambda g: cv2.Sobel(g, cv2.CV_64F, 1, 0, ksize=3),
            lambda g: cv2.Sobel(g, cv2.CV_64F, 0, 1, ksize=3),
            lambda g: cv2.Laplacian(g, cv2.CV_64F),
        ]:
            fi = fn(gray_img).flatten()
            fr = fn(gray_ref).flatten()
            filters_img.append(fi / (np.linalg.norm(fi) + 1e-8))
            filters_ref.append(fr / (np.linalg.norm(fr) + 1e-8))

        feat_img = np.stack(filters_img, axis=0)  # (3, N)
        feat_ref = np.stack(filters_ref, axis=0)

        gram_img = feat_img @ feat_img.T
        gram_ref = feat_ref @ feat_ref.T

        style_dist = float(np.mean((gram_img - gram_ref) ** 2))

        # Combined score
        return float(content_dist + 0.5 * style_dist)



"""Tests for vendi module."""

from ..conftest import _test_module_basics
from ayase.models import QualityMetrics


def test_vendi_basics():
    from ayase.modules.vendi import VendiModule
    _test_module_basics(VendiModule, "vendi")

def test_vendi_extract(video_sample):
    from ayase.modules.vendi import VendiModule
    m = VendiModule()
    feat = m.extract_features(video_sample)
    # May be None for non-video or missing deps
    assert video_sample is not None



"""DAVIS J&F — Video Object Segmentation Quality (DAVIS 2016).

Full-reference metric for evaluating video segmentation quality:
  J (Jaccard / IoU): region-based accuracy of predicted masks
  F (F-measure): contour-based accuracy of predicted masks

Expects reference segmentation masks. Uses heuristic based on
mask IoU and boundary F-measure computation.

davis_j — 0-1, higher = better (region IoU)
davis_f — 0-1, higher = better (boundary F-measure)
"""

import logging
import cv2
import numpy as np
from pathlib import Path
from typing import Optional

from ayase.models import Sample, QualityMetrics
from ayase.base_modules import ReferenceBasedModule

logger = logging.getLogger(__name__)


class DAVISJFModule(ReferenceBasedModule):
    name = "davis_jf"
    description = "DAVIS J&F video segmentation quality (FR, 2016)"
    metric_field = None  # We override process() to set both davis_j and davis_f
    default_config = {"subsample": 8, "boundary_threshold": 2}

    def __init__(self, config=None):
        super().__init__(config)
        self._model = None
        self.subsample = self.config.get("subsample", 8)
        self.boundary_threshold = self.config.get("boundary_threshold", 2)

    def setup(self) -> None:
        logger.info("DAVIS J&F module initialised (heuristic)")

    def compute_reference_score(self, sample_path: Path, reference_path: Path) -> Optional[float]:
        """Not used directly; process() is overridden instead."""
        return None

    def process(self, sample: Sample) -> Sample:
        """Override to set both davis_j and davis_f."""
        reference = getattr(sample, "reference_path", None)
        if reference is None:
            return sample

        if not isinstance(reference, Path):
            reference = Path(reference)
        if not reference.exists():
            return sample

        try:
            if str(sample.path).lower().endswith((".mp4", ".avi", ".mov", ".mkv", ".webm")):
                scores = self._score_video(str(sample.path), str(reference))
            else:
                scores = self._score_image(str(sample.path), str(reference))

            if scores is None:
                return sample

            if sample.quality_metrics is None:
                sample.quality_metrics = QualityMetrics()

            sample.quality_metrics.davis_j = scores["j"]
            sample.quality_metrics.davis_f = scores["f"]
            logger.debug(
                f"DAVIS J&F for {sample.path.name}: "
                f"J={scores['j']:.4f} F={scores['f']:.4f}"
            )
        except Exception as e:
            logger.warning(f"DAVIS J&F failed: {e}")
        return sample

    def _score_image(self, sample_p: str, ref_p: str) -> Optional[dict]:
        pred = cv2.imread(sample_p, cv2.IMREAD_GRAYSCALE)
        gt = cv2.imread(ref_p, cv2.IMREAD_GRAYSCALE)
        if pred is None or gt is None:
            return None

        h, w = gt.shape[:2]
        pred = cv2.resize(pred, (w, h))

        pred_mask = (pred > 127).astype(np.uint8)
        gt_mask = (gt > 127).astype(np.uint8)

        j = self._compute_jaccard(pred_mask, gt_mask)
        f = self._compute_boundary_f(pred_mask, gt_mask)
        return {"j": j, "f": f}

    def _score_video(self, sample_p: str, ref_p: str) -> Optional[dict]:
        cap_s = cv2.VideoCapture(sample_p)
        cap_r = cv2.VideoCapture(ref_p)
        try:
            total = min(
                int(cap_s.get(cv2.CAP_PROP_FRAME_COUNT)),
                int(cap_r.get(cv2.CAP_PROP_FRAME_COUNT)),
            )
            if total <= 0:
                return None

            indices = np.linspace(0, total - 1, min(self.subsample, total), dtype=int)
            j_scores = []
            f_scores = []

            for idx in indices:
                cap_s.set(cv2.CAP_PROP_POS_FRAMES, idx)
                cap_r.set(cv2.CAP_PROP_POS_FRAMES, idx)
                ret_s, frame_s = cap_s.read()
                ret_r, frame_r = cap_r.read()
                if not (ret_s and ret_r):
                    continue

                pred = cv2.cvtColor(frame_s, cv2.COLOR_BGR2GRAY) if frame_s.ndim == 3 else frame_s
                gt = cv2.cvtColor(frame_r, cv2.COLOR_BGR2GRAY) if frame_r.ndim == 3 else frame_r

                h, w = gt.shape[:2]
                pred = cv2.resize(pred, (w, h))

                pred_mask = (pred > 127).astype(np.uint8)
                gt_mask = (gt > 127).astype(np.uint8)

                j_scores.append(self._compute_jaccard(pred_mask, gt_mask))
                f_scores.append(self._compute_boundary_f(pred_mask, gt_mask))

            if not j_scores:
                return None

            return {
                "j": float(np.mean(j_scores)),
                "f": float(np.mean(f_scores)),
            }
        finally:
            cap_s.release()
            cap_r.release()

    def _compute_jaccard(self, pred: np.ndarray, gt: np.ndarray) -> float:
        """Compute Jaccard index (IoU) between binary masks."""
        intersection = np.logical_and(pred, gt).sum()
        union = np.logical_or(pred, gt).sum()
        if union == 0:
            return 1.0  # Both empty = perfect match
        return float(intersection / union)

    def _compute_boundary_f(self, pred: np.ndarray, gt: np.ndarray) -> float:
        """Compute boundary F-measure between binary masks.

        Extracts contours and computes precision/recall based on boundary
        pixel distance.
        """
        # Extract boundaries
        pred_boundary = self._get_boundary(pred)
        gt_boundary = self._get_boundary(gt)

        if pred_boundary.sum() == 0 and gt_boundary.sum() == 0:
            return 1.0  # Both have no boundaries
        if pred_boundary.sum() == 0 or gt_boundary.sum() == 0:
            return 0.0

        # Distance transform for tolerance matching
        from cv2 import distanceTransform, DIST_L2

        gt_dist = distanceTransform(1 - gt_boundary, DIST_L2, 3)
        pred_dist = distanceTransform(1 - pred_boundary, DIST_L2, 3)

        # Precision: predicted boundary pixels within threshold of GT boundary
        precision = float(
            np.sum(pred_boundary * (gt_dist <= self.boundary_threshold)) /
            (pred_boundary.sum() + 1e-8)
        )

        # Recall: GT boundary pixels within threshold of predicted boundary
        recall = float(
            np.sum(gt_boundary * (pred_dist <= self.boundary_threshold)) /
            (gt_boundary.sum() + 1e-8)
        )

        if precision + recall < 1e-8:
            return 0.0
        f_measure = 2 * precision * recall / (precision + recall)
        return float(np.clip(f_measure, 0.0, 1.0))

    def _get_boundary(self, mask: np.ndarray) -> np.ndarray:
        """Extract boundary pixels from a binary mask using morphological erosion."""
        kernel = np.ones((3, 3), np.uint8)
        eroded = cv2.erode(mask, kernel, iterations=1)
        boundary = (mask - eroded).astype(np.uint8)
        return boundary



"""AVQT — Apple Advanced Video Quality Tool.

Apple's perceptual video quality metric for content delivery.
Multi-scale perceptual comparison with human visual system modeling.

avqt_score — higher = better quality
"""

import logging
import subprocess
from pathlib import Path
from typing import Optional

import cv2
import numpy as np

from ayase.models import Sample, QualityMetrics
from ayase.base_modules import ReferenceBasedModule

logger = logging.getLogger(__name__)


class AVQTModule(ReferenceBasedModule):
    name = "avqt"
    description = "Apple AVQT perceptual video quality (full-reference)"
    metric_field = "avqt_score"
    default_config = {
        "subsample": 8,
    }

    def __init__(self, config=None):
        super().__init__(config)
        self.subsample = self.config.get("subsample", 8)
        self._model = None
        self._cli_available = False
        self._backend = "heuristic"

    def setup(self) -> None:
        # Tier 1: Try AVQT CLI tool
        try:
            result = subprocess.run(
                ["avqt", "--version"],
                capture_output=True,
                text=True,
                timeout=5,
            )
            if result.returncode == 0:
                self._cli_available = True
                self._backend = "cli"
                logger.info("AVQT (CLI) initialised")
                return
        except (FileNotFoundError, subprocess.TimeoutExpired, OSError):
            pass

        # Tier 2: Heuristic fallback
        self._backend = "heuristic"
        logger.info("AVQT (heuristic) initialised — install avqt CLI for full model")

    def compute_reference_score(
        self, sample_path: Path, reference_path: Path
    ) -> Optional[float]:
        if self._backend == "cli":
            return self._compute_cli(sample_path, reference_path)
        return self._compute_heuristic(sample_path, reference_path)

    def _compute_cli(
        self, sample_path: Path, reference_path: Path
    ) -> Optional[float]:
        """Run AVQT CLI tool."""
        try:
            result = subprocess.run(
                ["avqt", "--ref", str(reference_path), "--dis", str(sample_path)],
                capture_output=True,
                text=True,
                timeout=300,
            )
            if result.returncode == 0:
                # Parse score from output
                for line in result.stdout.strip().split("\n"):
                    line = line.strip()
                    if "score" in line.lower() or "avqt" in line.lower():
                        parts = line.split()
                        for part in reversed(parts):
                            try:
                                return float(np.clip(float(part), 0.0, 1.0))
                            except ValueError:
                                continue
            return None
        except (subprocess.TimeoutExpired, OSError) as e:
            logger.warning(f"AVQT CLI failed: {e}")
            return None

    def _read_frames(self, path: Path) -> list:
        """Read frames from video or image."""
        frames = []
        is_video = path.suffix.lower() in {
            ".mp4", ".avi", ".mov", ".mkv", ".webm", ".flv", ".wmv",
        }

        if is_video:
            cap = cv2.VideoCapture(str(path))
            try:
                total = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
                if total <= 0:
                    return frames
                indices = np.linspace(0, total - 1, min(self.subsample, total), dtype=int)
                for idx in indices:
                    cap.set(cv2.CAP_PROP_POS_FRAMES, idx)
                    ret, frame = cap.read()
                    if ret:
                        frames.append(frame)
            finally:
                cap.release()
        else:
            img = cv2.imread(str(path))
            if img is not None:
                frames.append(img)

        return frames

    def _compute_heuristic(
        self, sample_path: Path, reference_path: Path
    ) -> Optional[float]:
        """Heuristic: multi-scale perceptual comparison."""
        dist_frames = self._read_frames(sample_path)
        ref_frames = self._read_frames(reference_path)

        if not dist_frames or not ref_frames:
            return None

        n_frames = min(len(dist_frames), len(ref_frames))
        dist_frames = dist_frames[:n_frames]
        ref_frames = ref_frames[:n_frames]

        frame_scores = []
        for i in range(n_frames):
            dist_gray = cv2.cvtColor(dist_frames[i], cv2.COLOR_BGR2GRAY).astype(np.float64)
            ref_gray = cv2.cvtColor(ref_frames[i], cv2.COLOR_BGR2GRAY).astype(np.float64)

            # Resize if needed
            if dist_gray.shape != ref_gray.shape:
                dist_gray = cv2.resize(dist_gray, (ref_gray.shape[1], ref_gray.shape[0]))

            # Multi-scale comparison (3 scales)
            scale_scores = []
            d_cur, r_cur = dist_gray.copy(), ref_gray.copy()

            for s in range(3):
                if d_cur.shape[0] < 8 or d_cur.shape[1] < 8:
                    break

                # SSIM-like comparison at this scale
                c1 = (0.01 * 255) ** 2
                c2 = (0.03 * 255) ** 2

                mu_d = cv2.GaussianBlur(d_cur, (11, 11), 1.5)
                mu_r = cv2.GaussianBlur(r_cur, (11, 11), 1.5)
                sigma_d2 = cv2.GaussianBlur(d_cur ** 2, (11, 11), 1.5) - mu_d ** 2
                sigma_r2 = cv2.GaussianBlur(r_cur ** 2, (11, 11), 1.5) - mu_r ** 2
                sigma_dr = cv2.GaussianBlur(d_cur * r_cur, (11, 11), 1.5) - mu_d * mu_r

                ssim_map = ((2 * mu_d * mu_r + c1) * (2 * sigma_dr + c2)) / (
                    (mu_d ** 2 + mu_r ** 2 + c1) * (sigma_d2 + sigma_r2 + c2)
                )
                scale_scores.append(float(np.mean(ssim_map)))

                # Downsample
                if d_cur.shape[0] > 16 and d_cur.shape[1] > 16:
                    d_cur = cv2.pyrDown(d_cur)
                    r_cur = cv2.pyrDown(r_cur)

            if scale_scores:
                # Weight finer scales more (perceptual importance)
                weights = [0.5, 0.3, 0.2][: len(scale_scores)]
                weights = np.array(weights) / sum(weights)
                frame_scores.append(float(np.sum(np.array(scale_scores) * weights)))

        if not frame_scores:
            return None

        score = float(np.mean(frame_scores))
        return float(np.clip(score, 0.0, 1.0))



import math
from dataclasses import dataclass
from functools import partial
from typing import Callable, Optional

import torch


@dataclass
class InitArgs:
    use_gaussian: bool = True  # gaussian vs uniform
    coeff_std: Optional[float] = None  # std coeff multiplier
    no_init: bool = False


def get_init_fn(
    args: InitArgs, input_dim: int, init_depth: Optional[int]
) -> Callable[[torch.Tensor], torch.Tensor]:
    """
    Init functions.
    """
    if args.no_init:
        return lambda x: x

    # standard deviation
    std = 1 / math.sqrt(input_dim)
    std = std if args.coeff_std is None else (args.coeff_std * std)

    # rescale with depth
    if init_depth is not None:
        std = std / math.sqrt(2 * init_depth)

    # gaussian vs uniform
    if args.use_gaussian:
        return partial(
            torch.nn.init.trunc_normal_, mean=0.0, std=std, a=-3 * std, b=3 * std
        )
    else:
        bound = math.sqrt(3) * std  # ensure the standard deviation is `std`
        return partial(torch.nn.init.uniform_, a=-bound, b=bound)



import torch
import numpy as np
from PIL import Image
from typing import List, Union
import copy
import time
from decord import VideoReader, cpu
import os
# current_dir = os.path.dirname(os.path.abspath(__file__))
# current_dir = os.path.join(current_dir, "perceptionlm")
# print(f'current_dir {current_dir}')
import sys
import tempfile
import requests

# sys.path.append(current_dir)
from .perceptionlm.core.args import dataclass_from_dict
from .perceptionlm.core.transforms.image_transform import get_image_transform
from .perceptionlm.core.transforms.video_transform import get_video_transform
from .perceptionlm.apps.plm.generate import PackedCausalTransformerGeneratorArgs, PackedCausalTransformerGenerator, load_consolidated_model_and_tokenizer


PERCEPTION_LM_MODELS = {
    'perception-lm-1b': {
        'path': 'facebook/Perception-LM-1B',
        'image_size': 448,
        'max_video_frames': 32,
        'max_num_tiles': 36,
    },
    'perception-lm-3b': {
        'path': 'facebook/Perception-LM-3B',
        'image_size': 448,
        'max_video_frames': 32,
        'max_num_tiles': 36,
    },
    'perception-lm-8b': {
        'path': 'facebook/Perception-LM-8B',
        'image_size': 448,
        'max_video_frames': 32,
        'max_num_tiles': 36,
    },
}


class PerceptionLMModel:
    video_mode = "direct"
    allows_image = True
    
    def __init__(self,
                 model_name='perception-lm-8b',
                 device='cuda',
                 cache_dir=None):
        assert model_name in PERCEPTION_LM_MODELS, f"Model {model_name} not found in PERCEPTION_LM_MODELS"
        self.model_name = model_name
        self.device = device
        self.cache_dir = cache_dir
        self.model_info = PERCEPTION_LM_MODELS[model_name]
        self.load_model()

        self.tmp_files = []
        
    def load_model(self):
        model_path = self.model_info['path']
        self.model, self.tokenizer, self.config = load_consolidated_model_and_tokenizer(model_path)
        self.model.eval()

    def load_images(self, paths: List[str], num_tiles: int = 4) -> List[Union[torch.Tensor, List[torch.Tensor]]]:
        processed_data = []
        for path in paths:
            if path.lower().endswith(('.mp4', '.avi', '.mov', '.mkv')):  # Video file
                video_frames = self.load_video(path, num_tiles)
                processed_data.append(video_frames)
            elif path.lower().endswith('.npy'):  # NumPy file
                np_array = np.load(path)
                if np_array.ndim == 3:  # Single image
                    image = Image.fromarray(np_array.astype('uint8'), 'RGB')
                    transform = get_image_transform(
                        vision_input_type=("vanilla" if num_tiles == 1 else self.config.data.vision_input_type),
                        image_res=self.model.vision_model.image_size,
                        max_num_tiles=num_tiles,
                    )
                    image_tensor, _ = transform(image)
                    processed_data.append(image_tensor)
                elif np_array.ndim == 4:  # Multiple frames
                    frames = [Image.fromarray(frame.astype('uint8'), 'RGB') for frame in np_array]
                    transform = get_video_transform(
                        image_res=self.model.vision_model.image_size,
                    )
                    frames_tensor, _ = transform((frames, len(frames), None, None, None))
                    processed_data.append(frames_tensor)
                else:
                    raise ValueError(f"Unexpected shape for NumPy array in {path}")
            else:  # Regular image file
                image = Image.open(path).convert('RGB')
                transform = get_image_transform(
                    vision_input_type=("vanilla" if num_tiles == 1 else self.config.data.vision_input_type),
                    image_res=self.model.vision_model.image_size,
                    max_num_tiles=num_tiles,
                )
                image_tensor, _ = transform(image)
                processed_data.append(image_tensor)
        return processed_data

    def load_video(self, video_path, num_frames):
        if video_path.startswith(("http://", "https://")):
            tmpfile = None  

            try:

                response = requests.get(video_path, stream=True)
                response.raise_for_status()


                with tempfile.NamedTemporaryFile(delete=False, suffix=".mp4") as tmpfile:

                    self.tmp_files.append(tmpfile.name)
                    for chunk in response.iter_content(chunk_size=8192):
                        tmpfile.write(chunk)
                    
                    tmpfile.flush()
                    video_path = tmpfile.name
            
            except Exception as e:
                if tmpfile:
                    os.remove(tmpfile.name)
                    self.tmp_files.remove(tmpfile.name)
                
                
                print(f'Error when downloading video {video_path} \n\nError: {e}')

                raise
                
        vr = VideoReader(video_path, ctx=cpu(0))
        total_frame_num = len(vr)
        uniform_sampled_frames = np.linspace(0, total_frame_num - 1, num_frames, dtype=int)
        frame_idx = uniform_sampled_frames.tolist()
        video_frames = vr.get_batch(frame_idx).numpy()
        
        # Convert to PIL Image list
        frames = [Image.fromarray(frame) for frame in video_frames]
        # Apply video transform
        transform = get_video_transform(
            image_res=self.model.vision_model.image_size,
        )
        video_tensor, _ = transform((video_path, num_frames, None, None, None))
        return video_tensor

    def clear_temp_files(self):
        for file_name in self.tmp_files:
            try:
                os.remove(file_name)
            except OSError:
                pass
        self.tmp_files.clear()
    
    def forward(self,
                paths: List[str],
                texts: List[str],
                num_frames: int = 32,
                num_tiles: int = 36,
                question_template: str = "Does this image show \"{}\"? Answer the question with Yes or No",
                answer_template: str = "Yes") -> torch.Tensor:
        assert len(paths) == len(texts), "Number of paths and texts must match"
        
        questions = [question_template.format(text) for text in texts]
        answers = [answer_template]
        
        processed_data = self.load_images(paths, num_tiles)
        
        lm_probs = []
        for data, question, answer in zip(processed_data, questions, answers):
            # Determine if this is a video or image
            media_type = "video" if isinstance(data, torch.Tensor) and data.dim() > 3 else "image"
            
            # Create prompt
            prompts = [(question, data)]
            
            # Create generator
            gen_cfg = dataclass_from_dict(
                PackedCausalTransformerGeneratorArgs,
                {"temperature": 0.0, "top_p": None, "top_k": None},
                strict=False,
            )
            generator = PackedCausalTransformerGenerator(gen_cfg, self.model, self.tokenizer)
            
            # Run generation for scoring (force output to match expected answer)
            _, _, _, logits = generator.generate(prompts)#, force_output=answer)
            print(f'Vocabulary Length {len(logits[0].shape)}')

            ids, _ = self.tokenizer._tokenize_for_generation(question="Yes", media=data)
            yes_id = ids[0]
            print(f'Yes Token ID {yes_id}')

            # Use loglikelihood as probability score
            lm_prob = torch.exp(logits[yes_id]).item()
            lm_probs.append(lm_prob)
        
        self.clear_temp_files()
        return torch.tensor(lm_probs)
    
    def generate(self,
            images: List[str],
            texts: List[str],
            num_frames: int = 32,
            num_tiles: int = 36,
            temperature: float = 0.0,
            top_p: float = None,
            top_k: int = None,
            max_new_tokens: int = 256) -> List[str]:
        assert len(images) == len(texts), "Number of paths and texts must match"
        
        processed_data = self.load_images(images, num_tiles)
        generated_texts = []
        
        for data, prompt in zip(processed_data, texts):
            # Determine if this is a video or image
            media_type = "video" if isinstance(data, torch.Tensor) and data.dim() > 3 else "image"
            
            # Create prompt
            prompts = [(prompt, data)]
            print(prompt)
            # Create generator
            gen_cfg = dataclass_from_dict(
                PackedCausalTransformerGeneratorArgs,
                {"temperature": temperature, "top_p": top_p, "top_k": top_k},
                strict=False,
            )
            generator = PackedCausalTransformerGenerator(gen_cfg, self.model, self.tokenizer)
            
            # Run generation
            start_time = time.time() 
            generation, _, _ = generator.generate(prompts)
            end_time = time.time()
            print(generation)
            for gen in generation:
                generated_texts.append(gen)
            
        
        self.clear_temp_files()
                
        return generated_texts
    
    def batch_generate(self,
                      media_paths: List[str],
                      questions: List[str],
                      media_types: List[str],
                      num_frames: int = 32,
                      num_tiles: int = 36,
                      temperature: float = 0.0,
                      top_p: float = None,
                      top_k: int = None,
                      max_new_tokens: int = 256) -> List[str]:
        """
        Batch generation for multiple media files and questions
        """
        assert len(media_paths) == len(questions) == len(media_types), "Inputs must have the same length"
        
        prompts = []
        for path, question, media_type in zip(media_paths, questions, media_types):
            if media_type == "image":
                transform = get_image_transform(
                    vision_input_type=("vanilla" if num_tiles == 1 else self.config.data.vision_input_type),
                    image_res=self.model.vision_model.image_size,
                    max_num_tiles=num_tiles,
                )
                image = Image.open(path).convert("RGB")
                image, _ = transform(image)
                prompts.append((question, image))
            elif media_type == "video":
                transform = get_video_transform(
                    image_res=self.model.vision_model.image_size,
                )
                video_info = (path, num_frames, None, None, None)
                frames, _ = transform(video_info)
                prompts.append((question, frames))
            else:
                raise NotImplementedError(f"Media type {media_type} not supported")
        
        # Create generator
        gen_cfg = dataclass_from_dict(
            PackedCausalTransformerGeneratorArgs,
            {"temperature": temperature, "top_p": top_p, "top_k": top_k},
            strict=False,
        )
        generator = PackedCausalTransformerGenerator(gen_cfg, self.model, self.tokenizer)
        
        # Run generation
        generation, _, _ = generator.generate(prompts)
        
        return generation


from collections import OrderedDict

import torch


@torch.no_grad()
def update_ema(ema_model, model, decay=0.9999):
    """
    Step the EMA model towards the current model.
    """
    ema_params = OrderedDict(ema_model.named_parameters())
    model_params = OrderedDict(model.named_parameters())

    for name, param in model_params.items():
        # TODO: Consider applying only to params that require_grad to avoid small numerical changes of pos_embed
        ema_params[name].mul_(decay).add_(param.data, alpha=1 - decay)


def requires_grad(model, flag=True):
    """
    Set requires_grad flag for all parameters in a model.
    """
    for p in model.parameters():
        p.requires_grad = flag



from configs.data import *
from configs.model import *
# ========================= data ==========================
# NOTE The train_file will not be used during the evaluation
train_file = available_corpus["msrvtt_1k_test"] 
test_file = dict(msvd_ret_test=available_corpus["msvd_ret_test"])

test_types = ["msvd_ret_test"]


num_workers = 6

best_key = ["msvd_ret_test_match", "t2v_r1"]

# ========================= input ==========================
num_frames = 4
num_frames_test = 4
batch_size = 8
batch_size_test = 4
max_txt_l = 40

inputs = dict(
    image_res=224,
    video_input=dict(
        num_frames="${num_frames}",
        sample_type="rand",
        num_frames_test="${num_frames_test}",
        sample_type_test="middle",
        random_aug=False,
    ),
    max_txt_l=dict(image="${max_txt_l}", video="${max_txt_l}"),
    batch_size=dict(image="${batch_size}", video="${batch_size}"),
    batch_size_test=dict(image="${batch_size_test}", video="${batch_size_test}"),
)

# ========================= model ==========================
text_enc = "bert_large"
model = dict(
    model_cls="InternVideo2_Stage2",
    vision_encoder=dict(
        # backbone
        name="pretrain_internvideo2_1b_patch14_224",
        img_size=224, 
        num_frames="${num_frames}",
        tubelet_size=1,
        patch_size=14, 
        d_model=1408,
        clip_embed_dim=768,
        clip_teacher_embed_dim=3200,
        clip_teacher_final_dim=768,
        clip_norm_type='l2',
        clip_return_layer=6,
        clip_student_return_interval=1,
        pretrained='your_model_path/1B_pt.pth',
        use_checkpoint=True,
        checkpoint_num=40,
        use_flash_attn=True,
        use_fused_rmsnorm=True,
        use_fused_mlp=True,
        # clip teacher
        clip_teacher=None,
        clip_input_resolution=224,
        clip_teacher_return_interval=1,
        # mask
        video_mask_type="random",
        video_mask_ratio=0.8,
        image_mask_type="random",
        image_mask_ratio=0.5,
        sep_image_video_pos_embed=True,
        keep_temporal=False,
        only_mask=True
    ),
    text_encoder="${TextEncoders[${text_enc}]}",
    multimodal=dict(enable=True),
    embed_dim=512,
    temp=0.07,
    find_unused_parameters=False
)

criterion = dict(
    loss_weight=dict(
        vtc=1.0, 
        mlm=1.0, 
        vtm=1.0, 
        mvm=0.0,
        uta=0.0,
    ),  # 0: disabled.
    vtm_hard_neg=True,
    mlm_masking_prob=0.5,
    distill_final_features=True,
    clip_loss_ratio=[1., 1.]
)

optimizer = dict(
    opt="adamW",
    lr=1e-5,
    opt_betas=[0.9, 0.98],  # default
    weight_decay=0.05,
    max_grad_norm=3.,  # requires a positive float, use -1 to disable
    # use a different lr for some modules, e.g., larger lr for new modules
    different_lr=dict(enable=False, module_names=[], lr=1e-3),
)

scheduler = dict(sched="cosine", epochs=1, min_lr_multi=0.01, warmup_epochs=0.2)

evaluate = True
deep_fusion = False
evaluation = dict(
    eval_frame_ensemble="concat",  # [concat, max, mean, lse]
    eval_x_only=False,
    k_test=128,
    eval_offload=True,  # offload gpu tensors to cpu to save memory.
)

use_half_precision = True
use_bf16 = True

gradient_checkpointing = True # for text encoder
use_flash_sdp = False
use_mem_efficient_sdp = False and not use_flash_sdp
compile_model = False

# ========================= wandb ==========================
wandb = dict(
    enable=False,
    entity="opengvlab",  # username or team name to store the runs, see https://docs.wandb.ai/ref/python/init
    project="InternVideo2-Stage2",  # setup in your command line
)
dist_url = "env://"
device = "cuda"
mode = "pt"

# ========================= others ==========================
output_dir = None  # output dir
resume = False  # if True, load optimizer and scheduler states as well
debug = False
log_freq = 100
seed = 42

save_latest = False
auto_resume = True
jump_evaluate = False
pretrained_path = ""

deepspeed = dict(
    enable=True,
    stage=1,
)



import numpy as np
import matplotlib.pyplot as plt

def im2col(img, k, stride=1):
    # Parameters
    m, n = img.shape
    s0, s1 = img.strides
    nrows = m - k + 1
    ncols = n - k + 1
    shape = (k, k, nrows, ncols)
    arr_stride = (s0, s1, s0, s1)

    ret = np.lib.stride_tricks.as_strided(img, shape=shape, strides=arr_stride)
    return ret[:, :, ::stride, ::stride].reshape(k*k, -1)


def integral_image(x):
    M, N = x.shape
    int_x = np.zeros((M+1, N+1))
    int_x[1:, 1:] = np.cumsum(np.cumsum(x, 0), 1)
    return int_x


def moments(x, y, k, stride):
    kh = kw = k

    k_norm = k**2

    x_pad = np.pad(x, int((kh - stride)/2), mode='reflect')
    y_pad = np.pad(y, int((kw - stride)/2), mode='reflect')

    int_1_x = integral_image(x_pad)
    int_1_y = integral_image(y_pad)

    int_2_x = integral_image(x_pad*x_pad)
    int_2_y = integral_image(y_pad*y_pad)

    int_xy = integral_image(x_pad*y_pad)

    mu_x = (int_1_x[:-kh:stride, :-kw:stride] - int_1_x[:-kh:stride, kw::stride] - int_1_x[kh::stride, :-kw:stride] + int_1_x[kh::stride, kw::stride])/k_norm
    mu_y = (int_1_y[:-kh:stride, :-kw:stride] - int_1_y[:-kh:stride, kw::stride] - int_1_y[kh::stride, :-kw:stride] + int_1_y[kh::stride, kw::stride])/k_norm

    var_x = (int_2_x[:-kh:stride, :-kw:stride] - int_2_x[:-kh:stride, kw::stride] - int_2_x[kh::stride, :-kw:stride] + int_2_x[kh::stride, kw::stride])/k_norm - mu_x**2
    var_y = (int_2_y[:-kh:stride, :-kw:stride] - int_2_y[:-kh:stride, kw::stride] - int_2_y[kh::stride, :-kw:stride] + int_2_y[kh::stride, kw::stride])/k_norm - mu_y**2

    cov_xy = (int_xy[:-kh:stride, :-kw:stride] - int_xy[:-kh:stride, kw::stride] - int_xy[kh::stride, :-kw:stride] + int_xy[kh::stride, kw::stride])/k_norm - mu_x*mu_y

    mask_x = (var_x < 0)
    mask_y = (var_y < 0)

    var_x[mask_x] = 0
    var_y[mask_y] = 0

    cov_xy[mask_x + mask_y] = 0

    return (mu_x, mu_y, var_x, var_y, cov_xy)


def vif_gsm_model(pyr, subband_keys, M):
    tol = 1e-15
    s_all = []
    lamda_all = []

    for subband_key in subband_keys:
        y = pyr[subband_key]
        y_size = (int(y.shape[0]/M)*M, int(y.shape[1]/M)*M)
        y = y[:y_size[0], :y_size[1]]

        y_vecs = im2col(y, M, 1)
        cov = np.cov(y_vecs)
        lamda, V = np.linalg.eigh(cov)
        lamda[lamda < tol] = tol
        cov = V@np.diag(lamda)@V.T

        y_vecs = im2col(y, M, M)

        s = np.linalg.inv(cov)@y_vecs
        s = np.sum(s * y_vecs, 0)/(M*M)
        s = s.reshape((int(y_size[0]/M), int(y_size[1]/M)))

        s_all.append(s)
        lamda_all.append(lamda)

    return s_all, lamda_all


def vif_channel_est(pyr_ref, pyr_dist, subband_keys, M):
    tol = 1e-15
    g_all = []
    sigma_vsq_all = []

    for i, subband_key in enumerate(subband_keys):
        y_ref = pyr_ref[subband_key]
        y_dist = pyr_dist[subband_key]

        lev = int(np.ceil((i+1)/2))
        winsize = 2**lev + 1

        y_size = (int(y_ref.shape[0]/M)*M, int(y_ref.shape[1]/M)*M)
        y_ref = y_ref[:y_size[0], :y_size[1]]
        y_dist = y_dist[:y_size[0], :y_size[1]]

        mu_x, mu_y, var_x, var_y, cov_xy = moments(y_ref, y_dist, winsize, M)

        g = cov_xy / (var_x + tol)
        sigma_vsq = var_y - g*cov_xy

        g[var_x < tol] = 0
        sigma_vsq[var_x < tol] = var_y[var_x < tol]
        var_x[var_x < tol] = 0

        g[var_y < tol] = 0
        sigma_vsq[var_y < tol] = 0

        sigma_vsq[g < 0] = var_y[g < 0]
        g[g < 0] = 0

        sigma_vsq[sigma_vsq < tol] = tol

        g_all.append(g)
        sigma_vsq_all.append(sigma_vsq)

    return g_all, sigma_vsq_all


def vif(img_ref, img_dist,  wavelet='steerable', full=True):
    assert wavelet in ['steerable', 'haar', 'db2', 'bio2.2'], 'Invalid choice of wavelet'
    M = 3
    sigma_nsq = 0.1

    if wavelet == 'steerable':
        from pyrtools.pyramids import SteerablePyramidSpace as SPyr
        pyr_ref = SPyr(img_ref, 4, 5, 'reflect1').pyr_coeffs
        pyr_dist = SPyr(img_dist, 4, 5, 'reflect1').pyr_coeffs
        subband_keys = []
        for key in list(pyr_ref.keys())[1:-2:3]:
            subband_keys.append(key)
    else:
        from pywt import wavedec2
        ret_ref = wavedec2(img_ref, wavelet, 'reflect', 4)
        ret_dist = wavedec2(img_dist, wavelet, 'reflect', 4)
        pyr_ref = {}
        pyr_dist = {}
        subband_keys = []
        for i in range(4):
            pyr_ref[(3-i, 0)] = ret_ref[i+1][0]
            pyr_ref[(3-i, 1)] = ret_ref[i+1][1]
            pyr_dist[(3-i, 0)] = ret_dist[i+1][0]
            pyr_dist[(3-i, 1)] = ret_dist[i+1][1]
            subband_keys.append((3-i, 0))
            subband_keys.append((3-i, 1))
        pyr_ref[4] = ret_ref[0]
        pyr_dist[4] = ret_dist[0]

    subband_keys.reverse()
    n_subbands = len(subband_keys)

    [g_all, sigma_vsq_all] = vif_channel_est(pyr_ref, pyr_dist, subband_keys, M)

    [s_all, lamda_all] = vif_gsm_model(pyr_ref, subband_keys, M)

    # for index,subband in enumerate(subband_keys):
    #     s = s_all[index]
    #     y = pyr_ref[subband]
    #     # s1 = np.repeat(s,3,0)
    #     # s_rep = np.repeat(s1,3,1)

    #     # G = y/s_rep
    #     plt.imshow(y,'gray')
    #     plt.savefig('/home/josh-admin/hdr/qa/hdr_vmaf/python_vmaf/saved_pics/G_subband'+str(index)+'.png')
    #     # dump(G,'/home/josh-admin/hdr/qa/hdr_vmaf/python_vmaf/saved_pics/Glogit_subband'+str(index)+'.z')

    nums = np.zeros((n_subbands,))
    dens = np.zeros((n_subbands,))
    for i in range(n_subbands):
        g = g_all[i]
        sigma_vsq = sigma_vsq_all[i]
        s = s_all[i]
        lamda = lamda_all[i]

        n_eigs = len(lamda)

        lev = int(np.ceil((i+1)/2))
        winsize = 2**lev + 1
        offset = (winsize - 1)/2
        offset = int(np.ceil(offset/M))

        g = g[offset:-offset, offset:-offset]
        sigma_vsq = sigma_vsq[offset:-offset, offset:-offset]
        s = s[offset:-offset, offset:-offset]

        for j in range(n_eigs):
            nums[i] += np.mean(np.log(1 + g*g*s*lamda[j]/(sigma_vsq+sigma_nsq)))
            dens[i] += np.mean(np.log(1 + s*lamda[j]/sigma_nsq))

    if not full:
        return np.mean(nums + 1e-4)/np.mean(dens + 1e-4)
    else:
        return np.mean(nums + 1e-4)/np.mean(dens + 1e-4), (nums + 1e-4), (dens + 1e-4)


def vif_spatial(img_ref, img_dist, k=11, sigma_nsq=0.1, stride=1, full=False):
    x = img_ref.astype('float32')
    y = img_dist.astype('float32')

    mu_x, mu_y, var_x, var_y, cov_xy = moments(x, y, k, stride)

    g = cov_xy / (var_x + 1e-10)
    sv_sq = var_y - g * cov_xy

    g[var_x < 1e-10] = 0
    sv_sq[var_x < 1e-10] = var_y[var_x < 1e-10]
    var_x[var_x < 1e-10] = 0

    g[var_y < 1e-10] = 0
    sv_sq[var_y < 1e-10] = 0

    sv_sq[g < 0] = var_x[g < 0]
    g[g < 0] = 0
    sv_sq[sv_sq < 1e-10] = 1e-10

    vif_val = np.sum(np.log(1 + g**2 * var_x / (sv_sq + sigma_nsq)) + 1e-4)/np.sum(np.log(1 + var_x / sigma_nsq) + 1e-4)
    if (full):
        # vif_map = (np.log(1 + g**2 * var_x / (sv_sq + sigma_nsq)) + 1e-4)/(np.log(1 + var_x / sigma_nsq) + 1e-4)
        # return (vif_val, vif_map)
        return (np.sum(np.log(1 + g**2 * var_x / (sv_sq + sigma_nsq)) + 1e-4), np.sum(np.log(1 + var_x / sigma_nsq) + 1e-4), vif_val)
    else:
        return vif_val


def msvif_spatial(img_ref, img_dist, k=11, sigma_nsq=0.1, stride=1, full=False):
    x = img_ref.astype('float32')
    y = img_dist.astype('float32')

    n_levels = 5
    nums = np.ones((n_levels,))
    dens = np.ones((n_levels,))
    for i in range(n_levels-1):
        if np.min(x.shape) <= k:
            break
        nums[i], dens[i], _ = vif_spatial(x, y, k, sigma_nsq, stride, full=True)
        x = x[:(x.shape[0]//2)*2, :(x.shape[1]//2)*2]
        y = y[:(y.shape[0]//2)*2, :(y.shape[1]//2)*2]
        x = (x[::2, ::2] + x[1::2, ::2] + x[1::2, 1::2] + x[::2, 1::2])/4
        y = (y[::2, ::2] + y[1::2, ::2] + y[1::2, 1::2] + y[::2, 1::2])/4

    if np.min(x.shape) > k:
        nums[-1], dens[-1], _ = vif_spatial(x, y, k, sigma_nsq, stride, full=True)
    msvifval = np.sum(nums) / np.sum(dens)

    if full:
        return msvifval, nums, dens
    else:
        return msvifval



import os
import shutil
import socket


def has_slurm():
    """determine the system has slurm or not
    Returns: True if has else False.

    """
    return shutil.which("sbatch") is not None

def random_port():
    """random a unused port
    Returns: str

    """
    with socket.socket() as s:
        s.bind(("", 0))
        return s.getsockname()[1]

def runcmd(cmd):
    """run command

    Args:
        cmd (str): The command to run

    """
    os.system(cmd)



"""Tests for pc_psnr module."""

from ..conftest import _test_module_basics
from ayase.models import QualityMetrics


def test_pc_psnr_basics():
    from ayase.modules.pc_psnr import PCPSNRModule
    _test_module_basics(PCPSNRModule, "pc_psnr")

def test_pc_psnr_no_reference(image_sample):
    from ayase.modules.pc_psnr import PCPSNRModule
    m = PCPSNRModule()
    result = m.process(image_sample)
    assert result is image_sample



"""Tests for ImageReward module."""

import pytest

from ..conftest import _test_module_basics
from ayase.models import QualityMetrics


def test_image_reward_basics():
    from ayase.modules.image_reward import ImageRewardModule

    _test_module_basics(ImageRewardModule, "image_reward")


def test_image_reward_skip_no_caption(image_sample):
    """Module should skip gracefully when no caption is available."""
    from ayase.modules.image_reward import ImageRewardModule

    m = ImageRewardModule()
    # Don't call setup so _ml_available stays False
    result = m.process(image_sample)
    assert result is image_sample
    # No score should be set
    if result.quality_metrics:
        assert result.quality_metrics.image_reward_score is None


def test_image_reward_skip_no_ml(image_sample):
    """Module should skip gracefully when ML dependencies are missing."""
    from ayase.modules.image_reward import ImageRewardModule

    image_sample.quality_metrics = QualityMetrics()
    m = ImageRewardModule()
    # Don't call setup — _ml_available stays False
    result = m.process(image_sample)
    assert result is image_sample
    assert result.quality_metrics.image_reward_score is None


def test_image_reward_field_exists():
    """Verify image_reward_score field exists in QualityMetrics."""
    qm = QualityMetrics()
    assert hasattr(qm, "image_reward_score")
    assert qm.image_reward_score is None
    qm.image_reward_score = 1.5
    assert qm.image_reward_score == 1.5


def test_image_reward_field_group():
    """Verify image_reward_score is in the alignment group."""
    qm = QualityMetrics()
    groups = qm._FIELD_GROUPS
    assert groups.get("image_reward_score") == "alignment"
